import hashlib
import tarfile
import tempfile
from pathlib import Path
from typing import Union

import yaml

from ....client import HLClient
from ...dataset import Dataset, S3Files
from ...interfaces import IWriter

PathLike = Union[str, Path]


class AwsS3Writer(IWriter):
    format_name = "aws-s3"

    def __init__(
        self,
        dataset_id: int,
        data_files_cache_dir: PathLike,
        bucket_name: str,
        prefix: str,
        client: HLClient,
    ):
        self.dataset_id = dataset_id
        self.data_files_cache_dir = Path(data_files_cache_dir)
        self.bucket_name = bucket_name
        self.prefix = prefix
        self.client = client

    def write(
        self,
        dataset: Dataset,
    ):
        if self.prefix is None:
            prefix = Path(str(self.dataset_id))
        else:
            prefix = Path(f"{self.prefix}/{self.dataset_id}")

        def prepend_dataset_id(filename: str, dataset_id=self.dataset_id) -> str:
            filename = Path(filename)
            new_name = f"{dataset_id}_{filename.stem}{filename.suffix}"

            return str(filename.parent / new_name)

        idf = dataset.images_df
        id_to_new_id = {i: f"{self.dataset_id}_{i}" for i in idf.data_file_id}
        id_to_new_filename = {
            i: prepend_dataset_id(f) for i, f in zip(dataset.idf.data_file_id, dataset.idf.filename)
        }

        idf = dataset.idf.copy()
        idf["old_filename"] = idf.loc[:, "filename"]
        idf.loc[:, "filename"] = idf.data_file_id.map(id_to_new_filename)
        idf.loc[:, "data_file_id"] = idf.data_file_id.map(id_to_new_id)

        adf = dataset.adf.copy()
        adf.loc[:, "data_file_id"] = adf.data_file_id.map(id_to_new_id)

        tmp_dataset = Dataset(images_df=idf, annotations_df=adf)

        # Files will be added to the archive so to preserve the
        # path in the records.json:images:filename
        with tempfile.TemporaryDirectory() as tmp_dir:
            tmp_dir = Path("zzz-tmp")
            tmp_tar_path = Path(tmp_dir) / "files.tar"
            with tarfile.open(str(tmp_tar_path), "w") as tar:
                for old_filename, new_filename in zip(idf.old_filename, idf.filename):
                    file_location = self.data_files_cache_dir / old_filename
                    file_path_in_archive = Path("files") / new_filename

                    tar.add(str(file_location), arcname=str(file_path_in_archive))

            # Calculate MD5 sum of files archive
            with tmp_tar_path.open("rb") as f:
                files_md5 = hashlib.md5(f.read()).hexdigest()

            s3 = self.client.get_s3_client()
            files_s3_prefix = prefix / f"files_{self.dataset_id}_{files_md5}.tar"

            tmp_dataset.cloud_files_info = [
                S3Files(bucket_name=self.bucket_name, prefix=str(prefix), files=[files_s3_prefix.name])
            ]

            records_json_path = tmp_dir / "records.json"
            tmp_dataset.write_json(
                records_json_path,
            )
            ## Calculate MD5 sum of records.json
            with open(records_json_path, "rb") as f:
                records_md5 = hashlib.md5(f.read()).hexdigest()

            records_s3_prefix = prefix / f"records_{self.dataset_id}_{records_md5}.json"

            manifest = {"files": [files_s3_prefix.name], "records": [records_s3_prefix.name]}
            manifest_path = tmp_dir / "manifest.yaml"
            with manifest_path.open("w") as f:
                yaml.dump(manifest, f)

            manifest_s3_prefix = prefix / "manifest.yaml"

            print(f"Uploading {tmp_tar_path} to s3://{self.bucket_name}/{files_s3_prefix}")
            s3.upload_file(str(tmp_tar_path), self.bucket_name, str(files_s3_prefix))
            print(f"Uploading {records_json_path} to s3://{self.bucket_name}/{records_s3_prefix}")
            s3.upload_file(str(records_json_path), self.bucket_name, str(records_s3_prefix))
            print(f"Uploading {manifest_path} to s3://{self.bucket_name}/{manifest_s3_prefix}")
            s3.upload_file(str(manifest_path), self.bucket_name, str(manifest_s3_prefix))



# To Do
# ~~~~~
# - None, yet.
#
# Notes
# ~~~~~
# gst-launch-1.0 v4l2src device=/dev/video0 ! videoconvert ! autovideosink
#
# To Do
# ~~~~~
# - Support Mac OS X camera sources

import os

from aiko_services.elements.gstreamer import *

__all__ = ["VideoCameraReader"]

# -----------------------------------------------------------------------------

class VideoCameraReader:
  def __init__(self, input_devicepath, width, height):
    if not os.path.exists(input_devicepath):
      raise ValueError("Device does not exist: " + input_devicepath)

    Gst = gst_initialise()

    gst_launch_command = "v4l2src device={} ! videoflip video-direction=horiz ! videoconvert ! videorate ! appsink name=sink".format(input_devicepath)
    pipeline = Gst.parse_launch(gst_launch_command)

    sink = pipeline.get_by_name("sink")
    sink_caps = "video/x-raw, format={}, width={}, height={}, framerate={}".format(utilities.get_format(), width, height, "10/1")
    sink.set_property("caps", Gst.caps_from_string(sink_caps))

    self.video_reader = VideoReader(pipeline, sink)

  def queue_size(self):
    return self.video_reader.queue.qsize()

  def read_frame(self, timeout = None):
    return self.video_reader.read_frame(timeout)



from typing import Dict, List, Optional, Tuple

from aiko_services.elements.media import image_io

from highlighter.agent.capabilities import Capability
from highlighter.agent.capabilities.base_capability import StreamEvent
from highlighter.client.base_models.data_file import DataFile

__all__ = ["ImageOverlay", "ImageResize"]


class ImageResize(Capability):
    class DefaultStreamParameters(Capability.DefaultStreamParameters):
        width: int = 0  # If zero, will maintain aspect ratio with respect to height
        height: int = 0  # If zero, will maintain aspect ratio with respect to width

    @property
    def width(self) -> int:
        return self._get_parameter("width")[0]

    @property
    def height(self) -> int:
        return self._get_parameter("height")[0]

    def process_frame(self, data_files: List[DataFile]) -> Tuple[StreamEvent, Optional[Dict]]:
        data_files = [df.resize(width=self.width, height=self.height) for df in data_files]
        return StreamEvent.OKAY, {"data_files": data_files}


class ImageOverlay(image_io.ImageOverlay, Capability):

    def process_frame(self, stream, data_files, annotations):
        for df, annos in zip(data_files, annotations):
            df_anns = [a for a in annos if a.data_file_id == df.file_id]
            overlay = df.draw_annotations(df_anns)
            df.content = overlay
        return StreamEvent.OKAY, {"data_files": data_files}



from pathlib import Path
from typing import Any, List, Optional, Tuple
from uuid import UUID

import ultralytics

from highlighter.client.gql_client import HLClient
from highlighter.client.io import multithread_graphql_file_download
from highlighter.client.training_config import TrainingConfigType
from highlighter.core.const import OBJECT_CLASS_ATTRIBUTE_UUID
from highlighter.datasets.cropping import CropArgs
from highlighter.datasets.dataset import Dataset
from highlighter.datasets.formats.yolo.writer import YoloWriter
from highlighter.trainers import yolov11

__all__ = ["filter_dataset", "train", "evaluate"]

# Only used for classification.
# Modify this to change the cropping behaviour when training a
# a box classifier (a classifier trained on cropped detections)
CROP_ARGS = CropArgs(
    crop_rotated_rect=False,
    scale=None,
    pad=None,
)


def filter_dataset(dataset: Dataset) -> Tuple[Dataset, UUID, Optional[List[Any]]]:
    """Optionally add some code to filter the Highlighter Datasets as required.
    The YoloWriter will only use entities with both a pixel_location attribute
    and a 'category_attribute_id' attribute when converting to the Yolo dataset format.
    It will the unique values for the object_class attribute as the detection
    categories.

    For example, if you want to train a detector that finds Apples and Bananas,
    and your taxonomy looks like this:

        - object_class: Apple
        - object_class: Orange
        - object_class: Banana

    Then you may do something like this:

        adf = combined_ds.annotations_df
        ddf = combined_ds.data_files_df

        orange_entity_ids = adf[(adf.attribute_id == OBJECT_CLASS_ATTRIBUTE_UUID) &
                               (adf.value == "Orange")].entity_id.unique()

        # Filter out offending entities
        adf = adf[adf.entity_id.isin(orange_entity_ids)]

        # clean up images that are no longer needed
        ddf = ddf[ddf.data_file_id.isin(adf.data_file_id)]

        combined_ds.annotations_df = adf
    """
    # Only use attributes with this specific attribute_id as the
    # categories for your dataset
    category_attribute_id: UUID = OBJECT_CLASS_ATTRIBUTE_UUID

    # Optionally define a list of attribute values to use for your dataset.
    # If None, then use all unique attribute_values for the given
    # category_attribute_id as the categories
    categories: Optional[List[Any]] = None
    return dataset, category_attribute_id, categories


def train(training_config: TrainingConfigType, dataset: Dataset, training_run_dir: Path):

    yolo_dataset_dir = training_run_dir / "datasets"

    cfg = ultralytics.cfg.yaml_load("cfg.yaml")

    if cfg["task"] == "classify":
        crop_args = CROP_ARGS
    else:
        crop_args = None

    if not (yolo_dataset_dir / "data.yaml").exists():
        # Optionally filter dataset, see filter_dataset's doc str
        filtered_ds, category_attribute_id, categories = filter_dataset(dataset)

        # Download required images
        image_cache_dir = training_run_dir / "images"
        multithread_graphql_file_download(
            HLClient.get_client(),
            filtered_ds.data_files_df.data_file_id.values,
            image_cache_dir,
        )

        ddf = filtered_ds.data_files_df
        if any([Path(f).suffix.lower() == ".mp4" for f in ddf.filename.unique()]):
            print("Detected video dataset, interpolating data from keyframes")
            filtered_ds = filtered_ds.interpolate_from_key_frames(
                frame_save_dir=image_cache_dir,
                source_video_dir=image_cache_dir,
            )

        # Write dataset in yolo format
        writer = YoloWriter(
            output_dir=yolo_dataset_dir,
            image_cache_dir=image_cache_dir,
            category_attribute_id=category_attribute_id,
            categories=categories,
            task=cfg["task"],
            crop_args=crop_args,
        )
        writer.write(filtered_ds)

    return yolov11.train(cfg, training_run_dir, crop_args)


def evaluate():
    """ToDo"""
    pass



import json
from abc import ABC, abstractmethod
from dataclasses import asdict, dataclass, field
from datetime import datetime, timezone
from enum import Enum
from io import BytesIO
from pathlib import Path
from typing import IO, Any, Dict, List, Optional, Union
from uuid import UUID, uuid4

import fastavro
import numpy as np
from scipy.spatial.distance import pdist, squareform

from highlighter.client import HLClient
from highlighter.client import aws_s3 as hls3
from highlighter.client.base_models import Annotation, Entity, Observation

__all__ = [
    "ByteAvroEntityWriter",
    "EmbeddingAggregationMode",
    "EntityToAvroConverter",
    "FileAvroEntityWriter",
    "IEntityWriter",
    "S3AvroEntityWriter",
]


class EmbeddingAggregationMode(str, Enum):
    AVERAGE = "average"
    MOST_REPRESENTATIVE_WITHOUT_OUTLIERS = "most_representative_without_outliers"


def days_since_epoch(time):
    assert isinstance(time, datetime), f"created_at must be a datetime object, got {time}"
    return (time - datetime.fromtimestamp(0, timezone.utc)).days


def similarity_matrix(embeddings: np.ndarray) -> np.ndarray:
    """Get the full similarity matrix using 1 - cosine_distance.

    Args:
        embeddings: NxM numpy array where N is the number of embeddings to compare and M
        is the length of each vector

    Returns:
        an NxN array where matrix[r][c] is the distance from embeddings[i] to
        embeddings[j]

    """
    dist_upper_triangle = pdist(embeddings, metric="cosine")
    dist_matrix = squareform(dist_upper_triangle)
    similarity_matrix = 1 - dist_matrix
    np.fill_diagonal(similarity_matrix, 1)  # Set diagonal to 1
    return similarity_matrix


def average_embedding(embeddings: np.ndarray) -> np.ndarray:
    return embeddings.mean(axis=0)


def most_representative_without_outliers(embeddings: np.ndarray) -> np.ndarray:
    """Find the embedding that is closest to all other embeddings using
    1-cosine similarity as the distance metric

    If the number of embeddings is < 10 then we don't check for outliers
    """
    grid = similarity_matrix(embeddings)
    l = len(embeddings)  # noqa
    if l < 10:
        idx = np.argmax(np.sum(grid, axis=1))
        return embeddings[idx]

    grid = np.sort(grid, axis=1)
    bi, ti = l // 10, l - l // 10
    # print(l, bi, ti)
    grid = grid[:, bi:ti]
    idx = np.argmax(np.sum(grid, axis=1))
    return embeddings[idx]


def find_top_k_embeddings(embeddings: np.ndarray, k: int) -> np.ndarray:
    l = len(embeddings)  # noqa
    step = l // k
    result = np.empty((k, embeddings.shape[1]))
    for i in range(k):
        if i + 1 == k:
            result[i] = most_representative_without_outliers(embeddings[i * step :])
        else:
            result[i] = most_representative_without_outliers(embeddings[i * step : (i + 1) * step])
    return result


@dataclass
class TrackDatumSource:
    confidence: float
    frameId: int


@dataclass
class TrackObservation:
    entityId: str
    entityAttributeId: str
    entityAttributeEnumId: str
    entityDatumSource: TrackDatumSource
    value: Optional[Any]
    time: int  # days since epoch


@dataclass
class XY:
    x: float
    y: float


@dataclass
class Bounds:
    min: XY
    max: XY


@dataclass
class TrackDetection:
    frame_id: int
    geometry_type: int
    confidence: float
    bounds: Bounds


@dataclass
class Track:
    track_id: str = field(default_factory=lambda: str(uuid4()))
    detections: List[TrackDetection] = field(default_factory=list)
    eavts: List[TrackObservation] = field(default_factory=list)


@dataclass
class EntityTracks:
    id: str  # <- Entity.id
    object_class: str = ""
    tracks: List[Track] = field(default_factory=lambda: [Track()])
    embeddings: List[List[float]] = field(default_factory=lambda: [])

    def asdict(self):
        return asdict(self)


class EntityToAvroConverter:
    class CompressionCodec(str, Enum):
        NULL = "null"
        SNAPPY = "snappy"

    @staticmethod
    def _append_embeddings(
        observation: Observation,
        entity_tracks: EntityTracks,
    ) -> EntityTracks:
        entity_tracks.embeddings.append([int(i) for i in observation.value])
        return entity_tracks

    @staticmethod
    def _append_object_class(
        observation: Observation,
        entity_tracks: EntityTracks,
    ) -> EntityTracks:
        entity_tracks.object_class = str(observation.value)
        return entity_tracks

    @staticmethod
    def _append_enum(observation: Observation, entity_tracks: EntityTracks) -> EntityTracks:
        time = observation.occurred_at if hasattr(observation, "occurred_at") else observation.time
        entity_tracks.tracks[0].eavts.append(
            TrackObservation(
                entityId=str(observation.entity_id),
                entityAttributeId=str(observation.attribute_id),
                entityAttributeEnumId=str(observation.value),
                entityDatumSource=TrackDatumSource(
                    confidence=observation.datum_source.confidence,
                    frameId=observation.datum_source.frame_id,
                ),
                value=None,
                time=days_since_epoch(time),
            )
        )
        return entity_tracks

    APPEND_OBSERVATION_FNS = {
        "8beab557-8d83-4257-82d0-101341236c5b": _append_embeddings,
        "df10b67d-b476-4c4d-acc2-c1deb5a0e4f4": _append_object_class,
    }

    EmbeddingAggregationFn = {
        EmbeddingAggregationMode.AVERAGE: average_embedding,
        EmbeddingAggregationMode.MOST_REPRESENTATIVE_WITHOUT_OUTLIERS: most_representative_without_outliers,
    }

    @staticmethod
    def _annotation_to_track_frame(
        frame_id: int,
        annotation: Annotation,
    ) -> TrackDetection:
        x0, y0, x1, y1 = annotation.location.bounds
        confidence = annotation.datum_source.confidence
        return TrackDetection(
            frame_id=frame_id,
            bounds=Bounds(
                min=XY(x=float(x0), y=float(y0)),
                max=XY(x=float(x1), y=float(y1)),
            ),
            geometry_type=0,
            confidence=confidence,
        )

    def __init__(
        self,
        schema: Union[str, Path, List],
        compression_codec: CompressionCodec = CompressionCodec.NULL,
    ):
        self.compression_codec = compression_codec

        if isinstance(schema, (str, Path)):
            self._schema = json.load(Path(schema).open("r"))
        elif isinstance(schema, list):
            self._schema = schema
        else:
            raise ValueError(f"Invaid schema dict or schema path, got: {schema}")

    def to_records(self, entities_dict: Dict[UUID, List[Entity]]) -> List[EntityTracks]:
        records = []
        for entity_id, entities in entities_dict.items():
            if not entities:
                continue
            entity_tracks = EntityTracks(id=str(entity_id))
            for entity in entities:
                for anno in entity.annotations:

                    if (frame_id := anno.datum_source.frame_id) is None:
                        raise ValueError(
                            f"To use the {self.__class__.__name__} you "
                            "must specify 'frame_id' on the Annotation.datum_source"
                        )

                    for obs in anno.observations:
                        if (obs.datum_source.frame_id is not None) and (
                            obs.datum_source.frame_id != frame_id
                        ):
                            raise ValueError(
                                f"Observation and Annotation have conflicting frame_id, got: {obs} --- and --- {anno}"
                            )
                        entity_tracks = self._append_observation(obs, entity_tracks)

                    entity_tracks = self._append_annotation(anno, entity_tracks)

            records.append(entity_tracks.asdict())
        return records

    def write_records(self, records: List[EntityTracks], fp: Union[BytesIO, IO]):
        fastavro.writer(fp, self._schema, records, codec=self.compression_codec)
        fp.flush()
        return fp

    def convert(self, entities_dict: Dict[UUID, List[Entity]], fp: Union[BytesIO, IO]):
        records = self.to_records(entities_dict)
        return self.write_records(records, fp)

    def _append_observation(self, observation: Observation, entity_tracks: EntityTracks) -> EntityTracks:
        attribute_id = str(observation.attribute_id)
        append_entity_fn = self.APPEND_OBSERVATION_FNS.get(attribute_id, self._append_enum)
        entity_tracks = append_entity_fn(observation, entity_tracks)
        return entity_tracks

    def _append_annotation(self, annotation: Annotation, entity_tracks: EntityTracks) -> EntityTracks:
        entity_tracks.tracks[0].detections.append(
            self._annotation_to_track_frame(annotation.datum_source.frame_id, annotation)
        )
        return entity_tracks


class IEntityWriter(ABC):
    @abstractmethod
    def write(self, entities: Dict[UUID, List[Entity]]):
        pass


class ByteAvroEntityWriter(IEntityWriter):

    def __init__(
        self,
        schema: Union[str, Path, List],
        compression_codec: EntityToAvroConverter.CompressionCodec = EntityToAvroConverter.CompressionCodec.NULL,
    ):
        self.converter = EntityToAvroConverter(schema, compression_codec=compression_codec)

    def write(self, entities: Dict[UUID, List[Entity]]):
        avro_bytes = BytesIO()
        self.converter.convert(entities, avro_bytes)
        return bytes(avro_bytes.getbuffer())


class FileAvroEntityWriter(IEntityWriter):
    def __init__(
        self,
        schema: Union[str, Path, List],
        output_path: Union[str, Path],
        compression_codec: EntityToAvroConverter.CompressionCodec = EntityToAvroConverter.CompressionCodec.NULL,
    ):
        self.converter = EntityToAvroConverter(schema, compression_codec=compression_codec)
        self.output_path = Path(output_path)

    def write(self, entities: Dict[UUID, List[Entity]]):
        avro_bytes = BytesIO()
        self.converter.convert(entities, avro_bytes)
        with self.output_path.open("wb") as fp:
            fp.write(bytes(avro_bytes.getbuffer()))


class S3AvroEntityWriter(IEntityWriter):
    def __init__(
        self,
        schema: Union[str, Path, List],
        filename: str,
        client: Optional[HLClient] = None,
        compression_codec: EntityToAvroConverter.CompressionCodec = EntityToAvroConverter.CompressionCodec.NULL,
    ):
        self.converter = EntityToAvroConverter(schema, compression_codec=compression_codec)
        self.filename = filename

        if client is None:
            self.client = HLClient.from_env()
        else:
            self.client: HLClient = client

    def write(self, entities: Dict[UUID, List[Entity]]):
        avro_bytes = BytesIO()
        self.converter.convert(entities, avro_bytes)
        shrine_file = hls3.upload_file_to_s3_in_memory(
            self.client, bytes(avro_bytes.getbuffer()), self.filename, mimetype="application/avro"
        )
        return shrine_file



# Requires either ...
# - pip install PyGObject
# ... or ...
# - apt-get install python3-gi

from sys import platform

__all__ = [
    "get_format",
    "get_h264_decoder", "get_h264_encoder", "get_h264_encoder_options",
    "GStreamerError", "gst_initialise", "enable_opencv", "process_video"
]

# --------------------------------------------------------------------------- #

operating_system = "unknown"

if platform == "linux" or platform == "linux2":
  operating_system = "linux"
elif platform == "darwin":
  operating_system = "mac_os_x"

format = "RGB"
h264_decoder = { "linux": "avdec_h264", "mac_os_x": "avdec_h264" }
h264_encoder = { "linux": "x264enc",    "mac_os_x": "vtenc_h264" }

def get_format():
  return format

def get_h264_decoder():
  return h264_decoder[operating_system]

def get_h264_encoder():
  return h264_encoder[operating_system]

def get_h264_encoder_options():
  if operating_system != "linux": return ""

  return "tune=zerolatency speed-preset=ultrafast sliced-threads=true key-int-max=30"

# --------------------------------------------------------------------------- #

import gi

Gst = None
GstBase = None
GObject = None

class GStreamerError(Exception):
  def __init__(self, message):
    self.message = message

def gst_initialise(multiple_return_values=False):
  global Gst, GstBase, GObject

  if not Gst:
    gi.require_version('Gst', '1.0')
    gi.require_version('GstBase', '1.0')
    from gi.repository import Gst, GstBase, GObject
    Gst.init(None)

  if multiple_return_values:
    return Gst, GstBase, GObject
  else:
    return Gst

# --------------------------------------------------------------------------- #

cv2 = None

def enable_opencv():
  global cv2
  cv2 = None

  try:
    cv2 = __import__("cv2")
  except Exception: pass

  return cv2

def process_video(video_reader, video_writer):
  while True:
    frame = video_reader.read_frame(0.01)

    if frame != None:
      if "type" in frame and frame["type"] == "image":
#       print("Frame: ", frame["id"], video_reader.queue_size(), video_writer.queue_size())

        if cv2:
          image = cv2.cvtColor(frame["image"], cv2.COLOR_RGB2BGR)
          cv2.imshow("Video", image)
          if cv2.waitKey(1) & 0xff == ord('q'): break

          frame["image"] = cv2.putText(
            frame["image"], str(frame["id"]), (100, 400),
            cv2.FONT_HERSHEY_SIMPLEX, 2, (255,255,255), 2, cv2.LINE_AA)

      video_writer.write_frame(frame)

# --------------------------------------------------------------------------- #



from typing import Dict, List, Optional, Tuple
from uuid import UUID

from highlighter.agent.capabilities.base_capability import Capability, StreamEvent
from highlighter.client.base_models.data_file import DataFile
from highlighter.predictors.onnx_yolov8 import OnnxYoloV8 as Predictor

__all__ = ["OnnxYoloV8"]


class OnnxYoloV8(Capability):

    class DefaultStreamParameters(Capability.DefaultStreamParameters):
        onnx_file: str
        num_classes: int = 80
        class_lookup: Optional[Dict[int, Tuple[UUID, str]]] = None
        conf_thresh: float = 0.1
        nms_iou_thresh: float = 0.5
        is_absolute: bool = True

    def __init__(self, context, *args, **kwargs):
        context.get_implementation("PipelineElement").__init__(self, context, *args, **kwargs)
        self._predictor: Optional[Predictor] = None
        self._init_params: Optional[Tuple] = tuple()

    def start_stream(self, stream, stream_id, use_create_frame=True):
        super().start_stream(stream, stream_id, use_create_frame=use_create_frame)
        onnx_file, found = self._get_parameter("onnx_file")
        assert found

        num_classes, found = self._get_parameter("num_classes")
        assert found

        class_lookup, _ = self._get_parameter("class_lookup", default=None)
        conf_thresh, _ = self._get_parameter("conf_thresh")
        nms_iou_thresh, _ = self._get_parameter("nms_iou_thresh")
        is_absolute, _ = self._get_parameter("is_absolute")

        new_init_params = (onnx_file, num_classes, class_lookup, conf_thresh, nms_iou_thresh, is_absolute)

        if (self._predictor is None) or (new_init_params != self._init_params):
            self._predictor = Predictor(
                onnx_file,
                num_classes,
                class_lookup=class_lookup,
                conf_thresh=conf_thresh,
                nms_iou_thresh=nms_iou_thresh,
                is_absolute=is_absolute,
            )

        return StreamEvent.OKAY, {}

    def process_frame(self, stream, data_files: List[DataFile]) -> Tuple[StreamEvent, Dict]:
        annotations = self._predictor.predict(data_files)
        self.logger.debug(f"annotations: {annotations}")
        return StreamEvent.OKAY, {"annotations": annotations}



# Declaration order is based on the dependency on static references
#
# To Do
# ~~~~~
# - None, yet !

from .scheme_file import DataSchemeFile

from .scheme_tty import DataSchemeTTY

from .scheme_zmq import DataSchemeZMQ

from .audio_io import (
      AudioOutput
#     PE_AudioFilter, PE_AudioResampler,
#     PE_FFT, PE_GraphXY,
#     PE_MicrophonePA, PE_MicrophoneSD, PE_Speaker
)

from .elements import Mock, NoOp

from .image_io import (
    convert_image_to_numpy, convert_image_to_pil,
    ImageOutput, ImageOverlay, ImageOverlayFilter,
    ImageReadFile, ImageReadZMQ, ImageResize,
    ImageWriteFile, ImageWriteZMQ
)

from .text_io import (
    TextOutput, TextReadFile, TextReadZMQ,
    TextSample, TextTransform, TextWriteFile, TextWriteZMQ
)

from .video_io import VideoOutput, VideoReadFile, VideoShow, VideoWriteFile

from .webcam_io import VideoReadWebcam



from pathlib import Path

import click


def _to_pathlib_make_dir(ctx, param, value):
    if value is None:
        return value

    value = Path(value)
    value.mkdir(parents=True, exist_ok=True)
    return value


def _to_pathlib(ctx, param, value):
    if value is None:
        return value

    if isinstance(value, tuple):
        value = [Path(v) for v in value]
        return value

    if isinstance(value, str):
        value = Path(value)
        return value

    raise ValueError(f"Invalid path input: {v}")


class CommonOptions:
    annotations_dir = click.option(
        "-a",
        "--annotations-dir",
        type=click.Path(file_okay=False),
        required=True,
        help=("Directory to save the annotation files to. If it does not exist " "one will be created"),
        callback=_to_pathlib_make_dir,
    )

    data_file_dir = click.option(
        "-o",
        "--data-file-dir",
        type=click.Path(file_okay=False),
        required=False,
        default=None,
        help=("Directory to save the data_files to. Images with the same " "filename will be skipped"),
        callback=_to_pathlib_make_dir,
    )



from .const import *
from .geometry import *
from .hl_base_model import *
from .itertools import *
from .labeled_uuid import *
from .logging import *
from .pagination import *
from .visualize import *



#!/usr/bin/env python3
#
# Description
# ~~~~~~~~~~~
# Maintain list of available Services and stream (filtered) updates.
#
# Usage
# ~~~~~
# registrar [--primary]
#
#   TODO: --primary: Force take over of the primary registrar role
#
# NAMESPACE=aiko
# HOST=localhost
# PID=`ps ax | grep python | grep registrar | cut -d" " -f1`
# SID=0
# TAGS="(key1=value1 key2=value2)"
# TOPIC_PATH=$NAMESPACE/$HOST/$PID/$SID
#
# mosquitto_pub -t $TOPIC_PATH/in -m  \
#     "(add topic_prefix name protocol transport owner $TAGS)"
# mosquitto_pub -t $TOPIC_PATH/in -m  \
#     "(remove topic_prefix)"
# mosquitto_pub -t $TOPIC_PATH/in -m  \
#     "(share response * * * * $TAGS)"
#
# Notes
# ~~~~~
# Registrar subscribes to ...
# - TOPIC_REGISTRAR_BOOT:    (primary found ...), (primary absent)
# - {topic_path}/in:         (add ...), (history ...), (remove ...), (share ...)
# - {namespace}/+/+/+/state: (absent)
#
# Protocol
# ~~~~~~~~
# V2: 2023-02-21: Registrar supports "(history ...)" message
# V1: 2023-01-12: Registrar bootstrap message includes version number
#                 Renamed request message "query" to "share"
#                 Added Service name field
# V0: 2020-08-18: Initial version
#
# To Do
# ~~~~~
# * Define public API (available function calls) in the Registrar Interface
#
# * BUG: When ECProducer updates "service_count", need to int(services_count) !
#
# * BUG: Registrar won't become primary when there isn't another Registrar
#        and retained topic "aiko.TOPIC_REGISTRAR_BOOT" incorrectly indicates
#        that a primary Registrar is running and should say "(primary absent)"
#
# * BUG: If there are multiple secondaries, when the primary fails, then all
#        secondaries end up being primaries :(
#
# * Consider whether the Registrar should be an Actor instead of a Service ?
#
# * Secondary Registrar subscribe to primary Registrar and update "self.history"
#
# * Reimplement ServicesCache using ECProducer / ECConsumer
#
# - Allow Services to update ServiceDetails, e.g change tags on-the-fly
#
# * Create "Service" class, use everywhere and include "__str__()"
#   - Includes topic_path, protocol, transport, owner and tags
# - Implement Registrar as a sub-class of Category
#
# - Secondary Registrars should acquire Primary Registrar history
# - Secondary Registrars should periodically send a non-retained message to
#     the TOPIC_REGISTRAR_BOOT topic ... (secondary found ...) for Dashboard
# - Dashboard should show if ...
#   - TOPIC_REGISTRAR_BOOT indicates that there is no primary Registrar
#   - TOPIC_REGISTRAR_BOOT indicates that there is a primary Registrar,
#       but the primary is not responding to (history ...) or (share ...)
#   - TOPIC_REGISTRAR_BOOT indicates there are secondary Registrar(s)
#       and show their details
#
# - CLI: show [registrar_filter] ... show running Registrar state
# - CLI: kill service_filter ... terminate running Services
# - CLI: --primary ... Force take over of the primary registrar role
#
# - Primary Registrar supports discovery protocol for finding MQTT server, etc
# - Make this a sub-command of Aiko CLI
#
# - Handle MQTT restart
# - Handle MQTT stop and start on a different host
# - Handle if system crashes, then mosquitto doesn't get to send a LWT messages
#   for the Registrar leaving a stale reference to a Registrar now longer exists
#   If a new Registrar isn't started when the system restarts, then Aiko Clients
#   try to use the defunct Registrar
# - Consider the ability to add, change or remove a Service's details, e.g tags
# - When Service fails with LWT, publish timestamp on "topic_path/state"
#   - Maybe ProcessController should do this, rather than Registrar ?
# - Every Service persisted in MeemStore should have "uuid" Service tag
# - Document state and protocol
#   - Service state inspired by Meem life-cycle
# - Create registrar/protocol.py
# - Implement protocol.py and state_machine.py !
# - Primary and secondaries Registrars
#   - https://en.wikipedia.org/wiki/Raft_(algorithm)
#   - https://en.wikipedia.org/wiki/Conflict-free_replicated_data_type
#   - Eventual consistency and optimistic replication
# - Implement protocol matching similar to programming language interfaces
#     with inheritance
# - Add on_message_broker() handler to track MQTT connection status
#   - Events: "add", "remove", "timeout" (waiting for connection)
# - Add message handler for listening for other Registars ?
#     Add discovery protocol handler to keep a list of Registrars
#     This means the Aiko V2 framework should do the subscription automagically
#     - Find the primary registrar (if it exists ?)
#     - Find all other registars via "share"

import click
from collections import deque
import time

from aiko_services.main import *
from aiko_services.main.utilities import *

_LOGGER = aiko.logger(__name__)

_HISTORY_LIMIT_DEFAULT = 16
_HISTORY_RING_BUFFER_SIZE = 4096
_PRIMARY_SEARCH_TIMEOUT = 2.0  # seconds
_SERVICE_STATE_TOPIC = f"{get_namespace()}/+/+/+/state"
_TIME_STARTED = time.monotonic()

# --------------------------------------------------------------------------- #

class StateMachineModel():
    states = ["start", "primary_search", "secondary", "primary"]

    transitions = [
        {"source":
          "start", "trigger": "initialize", "dest": "primary_search"},
        {"source":
          "primary_search", "trigger": "primary_found", "dest": "secondary"},
        {"source":
          "primary_search", "trigger": "primary_promotion", "dest": "primary"},
        {"source":
          "primary", "trigger": "primary_failed", "dest": "primary_search"},
        {"source":
          "secondary", "trigger": "primary_failed", "dest": "primary_search"}
    ]

    def __init__(self, service):
        self.service = service

    def on_enter_primary_search(self, event_data):
#       parameters = event_data.kwargs.get("parameters", {})
        self.service.ec_producer.update("lifecycle", "primary_search")
        _LOGGER.debug("do primary_search add_timer")

# TODO: If oldest known secondary, then immediately become the primary
# TODO: Choose timer period as _PRIMARY_SEARCH_TIMEOUT +/- delta to avoid collisions
        event.add_timer_handler(
            self.primary_search_timer, _PRIMARY_SEARCH_TIMEOUT)

    def primary_search_timer(self):
        timer_valid = self.service.state_machine.get_state() == "primary_search"
        _LOGGER.debug(f"timer primary_search {timer_valid}")
        event.remove_timer_handler(self.primary_search_timer)
        if timer_valid:
            self.service.state_machine.transition("primary_promotion", None)

    def on_enter_secondary(self, event_data):
        self.service.ec_producer.update("lifecycle", "secondary")
        _LOGGER.debug("do enter_secondary")

    def on_enter_primary(self, event_data):
        self.service.ec_producer.update("lifecycle", "primary")
        _LOGGER.debug("do enter_primary")
        # Clear LWT, so this registrar doesn't receive another LWT on reconnect
        aiko.message.publish(aiko.TOPIC_REGISTRAR_BOOT, "", retain=True)
        payload_lwt = "(primary absent)"
        try:
            aiko.process.set_last_will_and_testament(  # May raise SystemError
                aiko.TOPIC_REGISTRAR_BOOT, payload_lwt, True)
            topic_path = self.service.topic_path
            time_started = self.service.time_started
            version = REGISTRAR_VERSION
            payload_out =  \
                f"(primary found {topic_path} {version} {time_started})"
            aiko.message.publish(
                aiko.TOPIC_REGISTRAR_BOOT, payload_out, retain=True)
        except SystemError as system_error:  # Probably MQTT server not running
            self.service.state_machine.transition("primary_failed", None)

# --------------------------------------------------------------------------- #

class Registrar(Service):
    Interface.default("Registrar", "aiko_services.main.registrar.RegistrarImpl")

class RegistrarImpl(Registrar):
    def __init__(self, context):
        context.get_implementation("Service").__init__(self, context)

        state_machine_model = StateMachineModel(self)
        self.state_machine = StateMachineOld(state_machine_model)

        self.history = deque(maxlen=_HISTORY_RING_BUFFER_SIZE)
        self.services = Services()

        self.share = {
            "lifecycle": "start",
            "log_level": get_log_level_name(_LOGGER),
            "source_file": f"v{REGISTRAR_VERSION}⇒ {__file__}",
            "service_count": 0
        }
        self.ec_producer = ECProducer(self, self.share)
        self.ec_producer.add_handler(self._ec_producer_change_handler)

        self.add_message_handler(
            self._service_state_handler, _SERVICE_STATE_TOPIC)
        self.add_message_handler(self._topic_in_handler, self.topic_in)
        self.set_registrar_handler(self._registrar_handler)

        self.state_machine.transition("initialize", None)

    def _ec_producer_change_handler(self, command, item_name, item_value):
        if item_name == "log_level":
            _LOGGER.setLevel(str(item_value).upper())

    def _registrar_handler(self, action, registrar):
        if action == "found":
            if self.state_machine.get_state() == "primary_search":
                self.state_machine.transition("primary_found", None)

        if action == "absent":
            if self.state_machine.get_state() == "primary_search":
                self.state_machine.transition("primary_promotion", None)
            else:
                self.services = Services()
                self.state_machine.transition("primary_failed", None)

    def _service_state_handler(self, _, topic, payload_in):
        command, parameters = parse(payload_in)
        if command == "absent" and topic.endswith("/state"):
            topic_path = topic[:-len("/state")]
            self._service_remove(topic_path)

    def _topic_in_handler(self, _, topic, payload_in):
        command, parameters = parse(payload_in)
        _LOGGER.debug(f"topic_in_handler(): {command}: {parameters}")

        if len(parameters) > 0:
            topic_path = parameters[0]
            if len(parameters) == 6:
                name = parameters[1]
                protocol = parameters[2]
                transport = parameters[3]
                owner = parameters[4]
                tags = parameters[5]

        if command == "add" and len(parameters) == 6:
            self._service_add(
                topic_path, name, protocol, transport, owner, tags, payload_in)

        if command == "remove" and len(parameters) == 1:
            self._service_remove(topic_path)

        if command == "history" and len(parameters) == 2:
            if parameters[1] == "*":
                count = _HISTORY_LIMIT_DEFAULT
            else:
                count = parse_int(parameters[1])
            if len(self.history) < count:
                count = len(self.history)

            payload_out = f"(item_count {count})"
            aiko.message.publish(topic_path, payload=payload_out)

            for service_details in self.history:
                if count < 1:
                    break
                service_tags = " ".join(service_details["tags"])
                payload_out =  "(add"                                \
                              f" {service_details['topic_path']}"    \
                              f" {service_details['name']}"          \
                              f" {service_details['protocol']}"      \
                              f" {service_details['transport']}"     \
                              f" {service_details['owner']}"         \
                              f" ({service_tags})"                   \
                              f" {service_details['time_add']}"      \
                              f" {service_details['time_remove']})"
                aiko.message.publish(topic_path, payload_out)
                count -= 1

        if command == "share" and len(parameters) == 6:
            filter = ServiceFilter("*", name, protocol, transport, owner, tags)
            services_out = self.services.filter_by_attributes(filter)

            payload_out = f"(item_count {services_out.count})"
            aiko.message.publish(topic_path, payload=payload_out)

            for service_details in services_out:
                service_tags = " ".join(service_details["tags"])
                payload_out =  "(add"                              \
                              f" {service_details['topic_path']}"  \
                              f" {service_details['name']}"        \
                              f" {service_details['protocol']}"    \
                              f" {service_details['transport']}"   \
                              f" {service_details['owner']}"       \
                              f" ({service_tags}))"
                aiko.message.publish(topic_path, payload_out)

            payload_out = f"(sync {topic_path})"
            aiko.message.publish(self.topic_out, payload_out)

    def _service_add(self,
        topic_path, name, protocol, transport, owner, tags, payload_out):

        if not self.services.get_service(topic_path):
            _LOGGER.debug(f"Service add: {topic_path}")

            service_details = {
                "topic_path": topic_path,
                "name": name,
                "protocol": protocol,
                "transport": transport,
                "owner": owner,
                "tags": tags,
                "time_add": time.monotonic(),
                "time_remove": 0
            }

            self.services.add_service(topic_path, service_details)
            self.ec_producer.update("service_count", self.services.count)

            aiko.message.publish(self.topic_out, payload_out)

    def _service_remove(self, topic_path):
        service_topic_path = ServiceTopicPath.parse(topic_path)
        if service_topic_path:
            if service_topic_path.service_id == "0":  # Process terminated
                process_topic_path, _ = ServiceTopicPath.topic_paths(topic_path)
                topic_paths =  \
                    self.services.get_process_services(process_topic_path)
            else:
                topic_paths = topic_path

            for topic_path in list(topic_paths):
                service_details = self.services.get_service(topic_path)
                if service_details:
                    _LOGGER.debug(f"Service remove: {topic_path}")

                    service_details["time_remove"] = time.monotonic()
                    self.history.appendleft(service_details)

                    self.services.remove_service(topic_path)
                    self.ec_producer.update(
                        "service_count", self.services.count)

                    payload_out = f"(remove {topic_path})"
                    aiko.message.publish(self.topic_out, payload_out)

# --------------------------------------------------------------------------- #

@click.command("main", help="Registrar Service")

def main():
    tags = ["ec=true"]  # TODO: Add ECProducer tag before add to Registrar
    init_args = service_args(
        REGISTRAR_SERVICE_TYPE, None, None, REGISTRAR_PROTOCOL, tags)
    registrar = compose_instance(RegistrarImpl, init_args)
    aiko.process.run(True)

if __name__ == "__main__":
    main()

# --------------------------------------------------------------------------- #



from datetime import UTC, datetime, timedelta
from enum import Enum
from typing import Any, Dict, List, Optional, Union
from uuid import UUID

from highlighter.client.base_models import SubmissionType
from highlighter.client.gql_client import HLClient
from highlighter.core.hl_base_model import GQLBaseModel

__all__ = [
    "update_task_status",
    "update_task",
    "lease_task",
    "lease_tasks_from_steps",
    "update_task_result",
]


class TaskStatus(str, Enum):
    PENDING = "PENDING"
    RUNNING = "RUNNING"
    FAILED = "FAILED"
    SUCCESS = "SUCCESS"

    @staticmethod
    def validate_str(s) -> bool:
        return s in [s.value for s in TaskStatus]


class UpdateTaskResultPayload(GQLBaseModel):
    submission: SubmissionType
    errors: List[Any]


class Case(GQLBaseModel):
    class CaseSubmission(GQLBaseModel):
        class DataFile(GQLBaseModel):
            uuid: str
            original_source_url: str
            file_url_original: str
            content_type: str

        id: str
        uuid: UUID
        data_files: List[DataFile]

    id: str
    latest_submission: CaseSubmission
    entity_id: Optional[str] = None


class Task(GQLBaseModel):
    id: str
    status: Optional[TaskStatus] = None
    case: Optional[Case] = None
    leased_until: Optional[datetime] = None
    parameters: Optional[Dict[str, Any]] = None


def update_task_result(
    client: HLClient,
    task_id: str,
    status: str,
    observations: List[Dict] = [],
    background_info_layer_file_data: Optional[Dict] = None,
):
    assert isinstance(status, TaskStatus) or TaskStatus.validate_str(status), f"Got: {status}"

    kwargs = {
        "id": task_id,
        "status": status,
        "eavtAttributes": observations,
    }
    if background_info_layer_file_data is not None:
        kwargs["backgroundInfoLayerFileData"] = background_info_layer_file_data

    result = client.updateTaskResultV2(return_type=UpdateTaskResultPayload, **kwargs)
    return result


def update_task_status(
    client: HLClient,
    task_id: str,
    status: Union[str, TaskStatus],
    message: Optional[str] = None,
):

    assert isinstance(status, TaskStatus) or TaskStatus.validate_str(status), f"Got: {status}"

    class UpdateTaskStatusResponse(GQLBaseModel):
        errors: List[Any]

    kwargs = {
        "id": task_id,
        "status": status,
    }
    if message is not None:
        kwargs["message"] = message

    response = client.update_task_status(return_type=UpdateTaskStatusResponse, **kwargs)
    if response.errors:
        raise ValueError(f"{response.errors}")

    return response


def update_task(
    client: HLClient,
    task_id: Union[UUID, str],
    status: Optional[Union[str, TaskStatus]] = None,
    leased_until: Optional[Union[datetime, str]] = None,
    lease_sec: Optional[int] = None,
    **kwargs,
) -> Task:

    if lease_sec:
        assert leased_until is None, "Cannot use both leased_until and lease_sec"
        leased_until = (datetime.now(UTC) + timedelta(seconds=lease_sec)).isoformat()

    if isinstance(leased_until, datetime):
        leased_until = leased_until.isoformat()

    class TaskResponse(GQLBaseModel):
        task: Task
        errors: Any

    kwargs = dict(
        id=str(task_id),
        status=status,
        leasedUntil=leased_until,
        **kwargs,
    )

    kwargs = {k: v for k, v in kwargs.items() if v is not None}

    response = client.updateTask(return_type=TaskResponse, **kwargs)

    if response.errors:
        raise ValueError(f"Errors: {response.errors}")

    return response.task


def lease_task(
    client: HLClient,
    task_id: Union[UUID, str],
    set_status_to: Optional[Union[str, TaskStatus]] = None,
    leased_until: Optional[Union[datetime, str]] = None,
    lease_sec: Optional[int] = None,
) -> Task:

    if lease_sec:
        assert leased_until is None, "Cannot use both leased_until and lease_sec"
        leased_until = (datetime.now(UTC) + timedelta(seconds=lease_sec)).isoformat()

    if isinstance(leased_until, datetime):
        leased_until = leased_until.isoformat()

    class TaskResponse(GQLBaseModel):
        task: Task
        errors: Any

    response = client.updateTask(
        return_type=TaskResponse, id=str(task_id), status=set_status_to, leasedUntil=leased_until
    )
    if response.errors:
        raise ValueError(f"Errors: {response.errors}")
    return response.task


def lease_tasks_from_steps(
    client: HLClient,
    step_ids: List[Union[UUID, str]],
    count: int = 1,
    filter_by_status: Optional[Union[str, TaskStatus]] = None,
    set_status_to: Optional[Union[str, TaskStatus]] = None,
    leased_until: Optional[Union[datetime, str]] = None,
    lease_sec: Optional[int] = None,
    filter_by_task_id: Optional[List[UUID]] = None,
) -> List[Task]:

    if filter_by_status:
        assert isinstance(filter_by_status, TaskStatus) or TaskStatus.validate_str(filter_by_status)

    if set_status_to:
        assert isinstance(set_status_to, TaskStatus) or TaskStatus.validate_str(set_status_to)

    if lease_sec:
        assert leased_until is None, "Cannot use both leased_until and lease_sec"
        leased_until = (datetime.now(UTC) + timedelta(seconds=lease_sec)).isoformat()

    if isinstance(leased_until, datetime):
        leased_until = leased_until.isoformat()

    class TaskResponse(GQLBaseModel):
        errors: List[Any]
        tasks: List[Task]

    response = client.leaseTasksFromSteps(
        return_type=TaskResponse,
        stepIds=[str(s) for s in step_ids],
        count=count,
        leasedUntil=leased_until,
    )

    if filter_by_status:
        tasks = [t for t in response.tasks if t.status == filter_by_status]
    else:
        tasks = response.tasks

    if filter_by_task_id:
        task_ids = []

        for i in filter_by_task_id:
            if isinstance(i, str):
                task_ids.append(i)
            elif isinstance(i, UUID):
                task_ids.append(str(i))
            else:
                raise ValueError()

        tasks = [t for t in tasks if t.id in task_ids]

    if set_status_to:
        for task in tasks:
            _ = update_task_status(client, task.id, set_status_to)
    return tasks


if __name__ == "__main__":
    x = lease_tasks_from_steps(
        HLClient.from_profile("haplomic"),
        ["87f57909-c8e0-412b-b117-59a170920460"],
        # leased_until="2024-09-24T05:05:55+0000",
        lease_sec=10,
        filter_by_status="PENDING",
        count=5,
    )

    print(x)



#!/usr/bin/env python3
#
# Aiko Service: Process Manager
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Create and destroy processes
#
# Usage
# ~~~~~
# ./process_manager.py  # run as an Aiko Service  # TO BE COMPLETED
#
# ./process_manager.py --example python
# ./process_manager.py --example shell
#
# To Do
# ~~~~~
# - Check out https://pypi.org/project/python-daemon
#
# - Use of threading versus multiprocessing
#   - https://docs.python.org/2/library/multiprocessing.html
#   - https://hackernoon.com/concurrent-programming-in-python-is-not-what-you-think-it-is-b6439c3f3e6a
#
# - Complete implementation as an Aiko Service
#   - Keep running, even when there are no processes to manage
#   - Share running processes
# - Handle process standard output and standard error

import click
import importlib
import os
import select
import time

from subprocess import Popen
from threading import Thread

from aiko_services.main import *

__all__ = ["ProcessManager"]


PROCESS_POLL_TIME = 0.2  # seconds
PROTOCOL_PROCESS_MANAGER = f"SERVICE_PROTOCOL_AIKO/process_manager:0"

# --------------------------------------------------------------------------- #

# poll = select.poll  # TODO: Is this required ?  Not available on Windows

class ProcessManager:
    def __init__(self, process_exit_handler=None):
        self.process_exit_handler = process_exit_handler
        self.processes = {}
        self.thread = None

    def __str__(self):
        output = ""
        for id, process_data in self.processes.items():
            pid = process_data["process"].pid
            command = process_data["command_line"][0]
            if output: output += "\n"
            output += f"{id}: {pid} {command}"
        return output

    def create(self, id, command, arguments=None):
        command_line = [command]
        file_extension = os.path.splitext(command)[-1]
        if file_extension not in [".py", ".sh"]:
            specification = None
            try:
                importlib.util.find_spec
            except AttributeError:  # Python < 3.4
                specification = importlib.find_loader(command)  # pylint: disable=deprecated-method
            else:
                specification = importlib.util.find_spec(command)
            if specification:
                command_line = [specification.origin]
                print(f'Module path resolved to: "{command_line}"')

        if arguments:
            command_line.extend(arguments)
        process = Popen(command_line, bufsize=0, shell=False)
        self.processes[id] = {
            "command_line": command_line,
            "process": process,
            "return_code": None
        }

        if not self.thread:
            self.thread = Thread(target=self.run)
            self.thread.start()

    def delete(self, id, terminate=True, kill=False):
        process_data = self.processes[id]
        del self.processes[id]
        process = process_data["process"]
        if terminate:
            process.terminate()
        if kill:
            process.kill()
        if self.process_exit_handler:
            self.process_exit_handler(id, process_data)

    def run(self):
        while len(self.processes):
            for id, process_data in list(self.processes.items()):
                process = process_data["process"]
                return_code = process.poll()
                if return_code is not None:
                    process_data["return_code"] = return_code
                    self.delete(id, terminate=False, kill=False)
            time.sleep(PROCESS_POLL_TIME)

def process_exit_handler_default(id, process_data):
    details = ""
    if process_data:
        command = process_data["command_line"][0]
        return_code = process_data["return_code"]
        details = f": {command} status: {return_code}"
    print(f"Exit process {id}" + details)

# --------------------------------------------------------------------------- #

def topic_in_handler(aiko, topic, payload_in):
    print(f"Message: {topic}: {payload_in}")

#   tokens = payload_in[1:-1].split()
#   if len(tokens) >= 1:
#       command = tokens[0]

#       if command == "task" and len(tokens) == 2:
#           operation = tokens[1]
#           if operation == "start":
#               if aks.get_parameter("publish_parameters") == "true":
#                   for parameter_name in aks_info.parameters[0]:
#                       parameter_value = aks.get_parameter(parameter_name)
#                       payload_out = f"{parameter_name}: {parameter_value}"
#                       aks_info.mqtt_client.publish(
#                           aks_info.TOPIC_OUT, payload_out)

#         payload_out = payload_in
#         mqtt_client.publish(aks_info.TOPIC_OUT, payload_in)

    return False

# --------------------------------------------------------------------------- #

def example_code(process_manager, example):
    if example == "ls":
        command_line = [ "ls", "-l" ]
        process = Popen(command_line, bufsize=0, shell=False)

    if example == "python":
        command = "./process_manager.py"
        arguments = [ "--example", "ls" ]
        process_manager.create("test_1", command, arguments)

    if example == "shell":
        command = "/bin/sh"
        arguments = [ "-c", "echo Start A; sleep  1; echo Stop A" ]
        process_manager.create("A", command, arguments)
        arguments = [ "-c", "echo Start B; sleep  2; echo Stop B" ]
        process_manager.create("B", command, arguments)
        arguments = [ "-c", "echo Start C; sleep 10; echo Stop C" ]
        process_manager.create("C", command, arguments)
        time.sleep(5)
        process_manager.delete("C")

# --------------------------------------------------------------------------- #

@click.command()
@click.option("--example", type=click.STRING, help="Run example")
@click.option("--tags", "-t", type=click.STRING, help="Aiko Service tags")

def main(example, tags):
    process_manager = ProcessManager(process_exit_handler_default)

    if example:
        example_code(process_manager, example)
    else:
        pass
    #   self.add_message_handler(self.topic_in_handler, self.topic_in)
    #   ServiceTags.parse_tags(tags)
    #   aiko.process.run(True)

if __name__ == "__main__":
    main()

# --------------------------------------------------------------------------- #



# To Do
# ~~~~~
# - Create an RTSP DataTarget, which is an RTSP Server producing images
#
# - Support "data_batch_size" resulting in a frame containing multiple images

import queue
from threading import Thread

import aiko_services as aiko
from aiko_services.elements.gstreamer import (
    get_format, gst_initialise, VideoReader)

__all__ = ["DataSchemeRTSP"]

# --------------------------------------------------------------------------- #
# parameter: "data_sources" provides the RTSP server details (incoming)
# - "data_sources" list should only contain a single entry
# - "(rtsp://hostname:port)"
# - "(rtsp://hostname:port/camera_channel)"
# - "(rtsp://username:password@hostname:port/camera_channel)"

class DataSchemeRTSP(aiko.DataScheme):
    def create_sources(self,
        stream, data_sources, frame_generator, use_create_frame=False):

        gst = gst_initialise()
        pipeline_element = self.pipeline_element
        rtsp_url = data_sources[0]
        self.share["rtsp_url"] = rtsp_url
        pipeline_element.logger.info(f"create_sources(): rtsp_url: {rtsp_url}")

        format, _ = pipeline_element.get_parameter("format", get_format())
        frame_rate, found = pipeline_element.get_parameter("frame_rate")
        if not found:
            diagnostic = 'Must provide "frame_rate" parameter'
            return aiko.StreamEvent.ERROR, {"diagnostic": diagnostic}
        resolution, found = pipeline_element.get_parameter("resolution")
        if not found:
            diagnostic = 'Must provide "resolution" parameter'
            return aiko.StreamEvent.ERROR, {"diagnostic": diagnostic}
        if isinstance(resolution, str):
            width, height = resolution.split("x")
            resolution = (int(width), int(height))

        gst_launch_command = f"rtspsrc location={rtsp_url} ! rtph264depay ! h264parse ! decodebin ! videoconvert ! videorate ! appsink name=sink"
        gst_pipeline = gst.parse_launch(gst_launch_command)
        sink = gst_pipeline.get_by_name("sink")
        sink_caps = f"video/x-raw, format={format}, width={width}, height={height}, framerate={frame_rate}"
        sink.set_property("caps", gst.caps_from_string(sink_caps))
        self.video_reader = VideoReader(gst_pipeline, sink)

        self.queue = queue.Queue()
        self.terminate = False

        pipeline_element.create_frames(stream, self.frame_generator, rate=0.0)
        return aiko.StreamEvent.OKAY, {}

    def create_targets(self, stream, data_targets):
        diagnostic = "DataSchemeRTSP does not implement create_targets()"
        return aiko.StreamEvent.ERROR, {"diagnostic": diagnostic}

    def destroy_sources(self, stream):
        self.terminate = True
        self.video_reader.stop()

    def destroy_targets(self, stream):
        diagnostic = "DataSchemeRTSP does not implement destroy_targets()"
        return aiko.StreamEvent.ERROR, {"diagnostic": diagnostic}

    def frame_generator(self, stream, frame_id):
        if self.terminate:
            return aiko.StreamEvent.STOP, {"diagnostic": "Terminated"}

        data_batch_size, _ = self.pipeline_element.get_parameter(
            "data_batch_size", default=1)
        data_batch_size = int(data_batch_size)

        frame = self.video_reader.read_frame(0.01)
        images = []
        if frame:
            if "type" in frame and frame["type"] == "image":
                images.append(frame["image"])
                stream.variables["timestamps"] = [frame["timestamp"]]

        if images:
            return aiko.StreamEvent.OKAY, {"images": images}
        else:
            return aiko.StreamEvent.NO_FRAME, {}

aiko.DataScheme.add_data_scheme("rtsp", DataSchemeRTSP)

# --------------------------------------------------------------------------- #



import io
import json
import sys
from datetime import datetime
from pathlib import Path
from typing import Dict, List, Optional, Tuple
from uuid import UUID

import numpy as np
from PIL import Image

from highlighter.client import HLJSONEncoder
from highlighter.client.base_models.data_file import DataFile
from highlighter.client.base_models.entity import Entity

from .base_capability import (
    Capability,
    ContextPipelineElement,
    DataSourceType,
    StreamEvent,
)

__all__ = [
    "EntityWriteFile",
    "WriteStdOut",
    "ImageWriteStdOut",
    "ImageWrite",
]


class BaseEntityWrite(Capability):

    class DefaultStreamParameters(Capability.DefaultStreamParameters):
        task_id: str = ""

    def __init__(self, context: ContextPipelineElement):
        super().__init__(context)
        self.frame_entities = dict()

    def get_task_id(self, stream) -> str:
        return stream.variables.get("task_id", None)

    def _get_source_file_location(self, stream):
        # ToDo: Find a better palce to put/get this from
        #       see also, ImageDataSource.process_frame
        source_info = stream.variables.get("source_info", {})
        source_file_location = source_info.get("source_file_location", None)
        if source_file_location is not None:
            return Path(source_file_location)
        return None

    def on_per_frame(self, stream, entities, data_file):
        pass

    def process_frame(
        self, stream, data_files: List[DataFile], entities: List[Dict[UUID, Entity]]
    ) -> Tuple[StreamEvent, dict]:
        for frame_entities, data_file in zip(entities, data_files):
            self.on_per_frame(stream, frame_entities, data_file)
        self.frame_entities[stream.frame_id] = entities
        return StreamEvent.OKAY, {}

    def on_stop_stream(self, stream, stream_id, entities):
        """Note this will not be called if you're calling `pipeline.process_frame`
        directly. Because this is called when a stream is stopped
        """
        pass

    def stop_stream(self, stream, stream_id) -> Tuple[StreamEvent, Optional[Dict]]:
        self.on_stop_stream(stream, stream_id, self.frame_entities)
        return StreamEvent.OKAY, {}


class EntityWriteFile(BaseEntityWrite):

    class DefaultStreamParameters(BaseEntityWrite.DefaultStreamParameters):
        """Can contain the following placeholders:

            {frame_id}
            {task_id}
            {timestamp}

        for example:
            per_frame_output_file = 'output_{frame_id}_{timestamp}.json'
        """

        per_frame_output_file: Optional[str] = None

        """Can contain the following placeholders:

            {task_id}
            {timestamp}
        """
        stop_stream_output_file: Optional[str] = None

    @property
    def per_frame_output_file(self) -> Optional[str]:
        value, _ = self._get_parameter("per_frame_output_file", None)
        return value

    @property
    def stop_stream_output_file(self) -> Optional[str]:
        value, _ = self._get_parameter("stop_stream_output_file", None)
        return value

    def _timestamp(self):
        return datetime.now().strftime("%Y%m%d%H%M%S%f")

    def get_per_frame_output_file_path(self, stream, data_file):
        task_id = stream.stream_id
        frame_id = data_file.media_frame_index

        return self.per_frame_output_file.format(
            frame_id=frame_id,
            task_id=task_id,
            timestamp=self._timestamp(),
        )

    def on_per_frame(self, stream, entities, data_file):
        if self.per_frame_output_file:
            output_file_path = self.get_per_frame_output_file_path(stream, data_file)
            Path(output_file_path).parent.mkdir(exist_ok=True, parents=True)
            output_str = json.dumps(entities, indent=2, sort_keys=True, cls=HLJSONEncoder)
            with open(output_file_path, "w") as f:
                f.write(output_str)
                self.logger.debug(f"{self.my_id()}: wrote {len(entities)} entities to {output_file_path} ")

    def get_on_stop_stream_output_file_path(self, stream):
        task_id = stream.variables.get("task_id", None)
        return self.stop_stream_output_file.format(
            stream_id=stream.stream_id,
            task_id=task_id,
        )

    def on_stop_stream(self, stream, stream_id, all_entities):
        if self.stop_stream_output_file:
            self.logger.debug(f"Writing stop_stream_output_file: {self.stop_stream_output_file}")
            output_file_path = self.get_on_stop_stream_output_file_path(stream)
            Path(output_file_path).parent.mkdir(parents=True, exist_ok=True)
            with open(output_file_path, "w") as f:
                f.write(json.dumps(all_entities, indent=2, sort_keys=True, cls=HLJSONEncoder))


class WriteStdOut(BaseEntityWrite):

    def on_per_frame(self, stream, entities, data_files):

        for entity in entities:
            output = {"frame_data": entity, "frame_id": stream.frame_id}
            print(json.dumps(output), file=sys.stdout, cls=HLJSONEncoder)


class ImageWriteStdOut(Capability):

    def process_frame(self, stream, data_files: List[DataSourceType]) -> Tuple[StreamEvent, Optional[Dict]]:
        image = data_files[0].content
        output_buffer = io.BytesIO()
        if isinstance(image, np.ndarray):
            Image.fromarray(image).save(output_buffer, format="PNG")
        elif isinstance(image, Image.Image):
            image.save(output_buffer, format="PNG")

        sys.stdout.buffer.write(output_buffer.getvalue())

        return StreamEvent.OKAY, {}


class ImageWrite(Capability):

    class DefaultStreamParameters(Capability.DefaultStreamParameters):
        output_dir: str

        # Can use placeholders, {file_id}, {media_frame_index}, {original_source_url}
        output_pattern: str

    @property
    def output_dir(self) -> str:
        value, _ = self._get_parameter("output_dir")
        return value

    @property
    def output_pattern(self) -> str:
        value, _ = self._get_parameter("output_pattern")
        return value

    def process_frame(self, stream, data_files: List[DataSourceType]) -> Tuple[StreamEvent, dict]:
        output_dir = Path(self.output_dir)
        output_dir.mkdir(exist_ok=True, parents=True)

        for df in data_files:
            filename = self.output_pattern.format(
                file_id=df.file_id,
                media_frame_index=df.media_frame_index,
                original_source_url=df.original_source_url,
            )
            dest = output_dir / filename
            image = df.content
            if isinstance(image, np.ndarray):
                Image.fromarray(image).save(dest)
            elif isinstance(image, Image.Image):
                image.save(dest)

        return StreamEvent.OKAY, {}



from typing import Dict, List, Optional, Sequence, Tuple, Union
from warnings import warn

import numpy as np
from PIL import Image
from shapely import affinity
from shapely import geometry as geom
from shapely import make_valid
from shapely.ops import unary_union

from .exceptions import require_package

try:
    import cv2
except ModuleNotFoundError as _:
    cv2 = None

__all__ = [
    "get_top_left_bottom_right_coordinates",
    "multipoly_from_mask",
    "multipolygon_from_coords",
    "polygon_from_coords",
    "polygon_from_left_top_width_height_coords",
    "polygon_from_mask",
    "polygon_from_tlbr",
    "try_make_polygon_valid_if_invalid",
    "resize_image",
]


def polygon_from_left_top_width_height_coords(coords: Union[Tuple[float, float, float, float], np.ndarray]):
    """Create a LocationAttributeValue for a box in the form [x,y,w,h]

    Args:
        coords: A sequence of box in the form [x,y,w,h]
    """
    x0, y0, w, h = coords
    x1 = x0 + w
    y1 = y0 + h
    _coords = [[x0, y0], [x1, y0], [x1, y1], [x0, y1], [x0, y0]]
    return geom.Polygon(_coords)


def try_make_polygon_valid_if_invalid(shape):
    if shape.is_valid:
        return shape

    valid_shape = make_valid(shape)
    if not isinstance(valid_shape, (geom.MultiPolygon, geom.Polygon)):
        raise ValueError(f"Invalid Polygon/MultiPolygon {shape}")
    return valid_shape


def polygon_from_coords(
    coords: Sequence[
        Union[Tuple[float, float], Tuple[float, float, float], np.ndarray, Sequence[geom.Point]]
    ],
    fix_invalid_polygons: bool = False,
):
    """
    Args:
        coords: A sequence of (x, y [,z]) numeric coordinate pairs or triples, or
        an array-like with shape (N, 2) or (N, 3).
        Also can be a sequence of Point objects.
    """
    if len(coords) < 3:
        raise ValueError(f"Polygon must have at least 3 coordinates: {coords}")

    value = geom.Polygon(coords)
    if fix_invalid_polygons:
        value = try_make_polygon_valid_if_invalid(value)
    return value


def polygon_from_tlbr(x: Tuple[int, int, int, int]) -> geom.Polygon:
    """
    from top left bottom right format
    """
    top_left = x[0], x[1]
    top_right = x[2], x[1]
    bottom_right = x[2], x[3]
    bottom_left = x[0], x[3]
    return geom.Polygon(np.array([top_left, top_right, bottom_right, bottom_left]))


def multipolygon_from_coords(
    coords: Sequence[Sequence[Tuple[float, float]]],
    fix_invalid_polygons: bool = False,
):
    """
    Args:
        coords: A nested sequence of (x, y) numeric coordinate pairs, or
        an array-like with shape (N, 2).
    """
    shapes = []
    for poly_xys in coords:
        shape = geom.Polygon(poly_xys)
        if fix_invalid_polygons:
            shape = try_make_polygon_valid_if_invalid(shape)
        shapes.append(shape)

    if len(shapes) == 1:
        value = shapes[0]
    else:
        value = unary_union(shapes)
    return value


def get_top_left_bottom_right_coordinates(
    value: geom.Polygon, scale: float = 1.0, scale_about_origin: bool = True, pad: int = 0
) -> Tuple[int, int, int, int]:
    """
    to top left bottom right format
    """
    bounds: geom.Polygon = geom.box(*value.bounds)

    if scale_about_origin:
        bounds = affinity.scale(bounds, xfact=scale, yfact=scale, origin=(0, 0, 0))
    else:
        bounds = affinity.scale(bounds, xfact=scale, yfact=scale, origin="center")

    bounds = tuple([int(i) for i in bounds.buffer(pad, join_style="bevel").bounds])
    return bounds


@require_package(cv2, "cv2", "opencv")
def polygon_from_mask(
    mask: np.ndarray,
    approx_level: float = 0.01,
    dilate_args: Optional[Dict] = {"kernel": (5, 5), "iterations": 7},
) -> Optional[np.ndarray]:
    """Convert a boolean mask into an nparray of coordinates representing
    a polygon [x0, y0, ... xn, yn]

      coords run anti-clockwise
      coords are approximate
    """
    if dilate_args is not None:

        # Define the dilation kernel
        kernel = np.ones(dilate_args["kernel"], np.uint8)
        mask_as_uint = mask.astype(np.uint8) * 255

        # Dilate the blurred mask
        _mask = cv2.dilate(mask_as_uint, kernel, iterations=dilate_args["iterations"])
    else:
        _mask = mask.astype(np.uint8) * 255

    # JANK OPEN CV VERSION CHECK to account for different cv2.findContours
    # returned tuple size
    contours_tuple = cv2.findContours(_mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)

    if len(contours_tuple) == 2:
        contours, _ = contours_tuple
    else:
        _, contours, _ = contours_tuple

    if len(contours) > 0:
        cnt = np.concatenate(contours, axis=0)
    else:
        warn("Could not convert mask to polygon points, return None")
        return None

    # FIXME Don't convexify by default?

    if len(contours) > 1:
        # See if I can reduce the number of points:
        cnt = cv2.convexHull(cnt)

    epsilon = approx_level * cv2.arcLength(cnt, True)
    cnt = cv2.approxPolyDP(cnt, epsilon, True)

    if cnt.shape[0] < 3:
        warn("Mask results in invalid polygon points, return None")
        return None

    return np.squeeze(cnt).flatten()


@require_package(cv2, "cv2", "opencv")
def multipoly_from_mask(
    mask: np.ndarray,
    dilate_args: Optional[Dict] = {"kernel": (5, 5), "iterations": 7},
) -> Optional[List[Tuple[int, int]]]:

    if dilate_args is not None:

        # Define the dilation kernel
        kernel = np.ones(dilate_args["kernel"], np.uint8)
        mask_as_uint = mask.astype(np.uint8) * 255

        # Dilate the blurred mask
        _mask = cv2.dilate(mask_as_uint, kernel, iterations=dilate_args["iterations"])
    else:
        _mask = mask.astype(np.uint8) * 255

    # JANK OPEN CV VERSION CHECK to account for different cv2.findContours
    # returned tuple size
    contours_tuple = cv2.findContours(_mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)

    if len(contours_tuple) == 2:
        contours, _ = contours_tuple
    else:
        _, contours, _ = contours_tuple

    approx = [
        np.squeeze(cv2.approxPolyDP(cnt, 0.01 * cv2.arcLength(cnt, True), True)).reshape((-1, 2))
        for cnt in contours
    ]
    # it's not a valid polygon if it has less than 3 vertices
    approx = [coords for coords in approx if len(coords) > 2]

    if len(approx) == 0:
        return None

    return approx


def resize_image(image: Union[np.ndarray, Image.Image], width: int = 0, height: int = 0):
    """
    Resize an image while maintaining the aspect ratio.

    Args:
        image (numpy.ndarray or PIL.Image.Image): The input image.
        width (int): The desired width (0 to auto-adjust based on height).
        height (int): The desired height (0 to auto-adjust based on width).

    Returns:
        numpy.ndarray or PIL.Image.Image: The resized image.
    """
    if cv2 is None:
        raise OptionalPackageMissingError("cv2", "opencv")

    if isinstance(image, np.ndarray):  # OpenCV (numpy array)
        h, w = image.shape[:2]
    elif isinstance(image, Image.Image):  # PIL Image
        w, h = image.size
    else:
        raise TypeError("Unsupported image type. Use numpy.ndarray or PIL.Image.")

    # Maintain aspect ratio if height or width is zero
    if width == 0 and height > 0:
        scale = height / h
        width = int(w * scale)
    elif height == 0 and width > 0:
        scale = width / w
        height = int(h * scale)
    elif width == 0 and height == 0:
        raise ValueError("Both width and height cannot be zero.")

    # Resize with the appropriate library
    if isinstance(image, np.ndarray):
        return cv2.resize(image, (width, height), interpolation=cv2.INTER_AREA)
    else:
        return image.resize((width, height), Image.Resampling.LANCZOS)



import json
import sys
from pathlib import Path
from queue import Queue
from urllib.parse import urlparse
from uuid import uuid4

import aiko_services as aiko
import click
from aiko_services.main import DEFAULT_STREAM_ID

from highlighter.agent.agent import (
    PIPELINE_TIMEOUT_SECONDS,
    TASK_LEASE_DURATION_SECONDS,
    HLAgent,
    SExpression,
    set_mock_aiko_messager,
)
from highlighter.client.agents import create_agent_token, create_machine_agent_version
from highlighter.client.base_models.data_file import DataFile
from highlighter.client.tasks import lease_task
from highlighter.core import HL_AGENT_QUEUE_RESPONSE_MAX_SIZE


@click.group("agent")
@click.pass_context
def agent_group(ctx):
    pass


# ToDo: Now that I am not accepting cli passed stream params. Is this needed,
# or is it just handled by the aiko pipeline
def parse_stream_parameters(agent, agent_definition) -> dict:

    stream_parameters = {}
    for node in agent.pipeline_graph.nodes():
        node_name = node.name

        # If default_stream_parameters is not present, we are likely
        # working with an aiko.PipelineElement. In this case, we can assume
        # that parameter validation is handled manually.
        if not hasattr(node.element, "default_stream_parameters"):
            continue

        # Start with in code params
        default_stream_parameters = node.element.default_stream_parameters()

        # Overwite with global pipeline definition params
        global_pipeline_definition_params = {
            k: v
            for k, v in agent_definition.parameters
            if k in node.element.DefaultStreamParameters.model_fields
        }
        default_stream_parameters.update(global_pipeline_definition_params)

        # Overwite with per element pipeline definition paras
        element_definition = [e for e in agent_definition.elements if e.name == node_name][0]
        pipeline_element_definition_params = {
            k.replace(f"{node_name}.", ""): v
            for k, v in element_definition.parameters.items()
            if k.replace(f"{node_name}.", "") in node.element.DefaultStreamParameters.model_fields
        }
        node.element.parameters = pipeline_element_definition_params
        default_stream_parameters.update(pipeline_element_definition_params)

        ele_stream_parameters = node.element.DefaultStreamParameters(**default_stream_parameters).model_dump()

        stream_parameters.update(
            {f"{element_definition.name}.{k}": v for k, v in ele_stream_parameters.items()}
        )

    return stream_parameters


def _reading_raw_data_from_stdin_buffer(input_data, expect_filepaths, seperator):
    return (input_data == "--") and (not sys.stdin.isatty() and (not expect_filepaths))


def _is_url(p):
    return all([urlparse(p), urlparse(p).netloc])


def _read_filepaths(input_data, seperator, encoding):
    """Should this belong in the HLFileDataScheme"""
    if input_data == "--":
        # Read raw bytes from stdin
        byte_input = sys.stdin.buffer.read()
        # Decode bytes to string using specified encoding
        text_input = byte_input.decode(encoding)
        # Split on separator and yield non-empty paths

        inputs = text_input.split(seperator)
    else:
        inputs = input_data

    # Take the first scheme and assume and assume all future schemes are the same
    scheme = None

    sources = []
    for path_url in inputs:
        path_url = path_url.strip()

        if Path(path_url).exists():  # Skip empty strings
            if scheme is None:
                scheme = "file"
            elif scheme != "file":
                raise ValueError("All schemes must be the same expected file")
            sources.append(f"file://{path_url}")
        elif _is_url(path_url):
            if scheme is None:
                scheme = "hlhttp"
            elif scheme != "hlhttp":
                raise ValueError("All schemes must be the same expected hlhttp")
            sources.append(f"hlhttp://{path_url}")
        else:
            raise NotImplementedError()

    assert len(sources) > 0
    return sources


def _set_agent_data_source_stream_parameters(agent, data_sources, stream_parameters):

    data_source_capabilities = agent.get_data_source_capabilities()
    if len(data_source_capabilities) == 1:
        data_source_capability_name = data_source_capabilities[0].name
    elif len(data_source_capabilities) > 1:
        raise NotImplementedError(
            f"hl agent start cannot yet support Agents with multiple DataSource Capabilities, got: {data_source_capabilities}"
        )
    else:
        raise NotImplementedError("hl agent start cannot yet support Agents with no DataSource Capabilities")

    input_name = f"{data_source_capability_name}.data_sources"
    stream_parameters[input_name] = data_sources
    return stream_parameters


def _data_sources_are_specified_in_agent_definition(agent):
    data_source_capabilities = agent.get_data_source_capabilities()
    for data_source_capability in data_source_capabilities:
        if data_source_capability.element.definition.parameters.get("data_sources"):
            return True
    return False


def process_data_sources(agent, stream_id, data_sources, _hlagent_cli_runner_queue_response):
    stream_parameters: dict = parse_stream_parameters(
        agent.pipeline,
        agent.pipeline_definition,
    )

    if data_sources:
        data_source_sexp = SExpression.encode(None, data_sources)
        stream_parameters = _set_agent_data_source_stream_parameters(
            agent, data_source_sexp, stream_parameters
        )

    queue_response = Queue(maxsize=HL_AGENT_QUEUE_RESPONSE_MAX_SIZE)
    agent.pipeline.create_stream(stream_id, parameters=stream_parameters, queue_response=queue_response)
    while True:
        stream_event, result = queue_response.get(timeout=PIPELINE_TIMEOUT_SECONDS)
        agent.pipeline.logger.debug(f"queue_response size: {queue_response.qsize()}")

        # This should only run if running via the HLAgentCliRunner
        if _hlagent_cli_runner_queue_response is not None:
            _hlagent_cli_runner_queue_response.put((stream_event, result))
            if _hlagent_cli_runner_queue_response.qsize() >= HL_AGENT_QUEUE_RESPONSE_MAX_SIZE:
                raise ValueError(
                    f"_hlagent_cli_runner_queue_response size is > {HL_AGENT_QUEUE_RESPONSE_MAX_SIZE}"
                )

        if stream_event["state"] in [aiko.StreamEvent.STOP, aiko.StreamEvent.ERROR]:
            break


def loop_over_process_frame(agent, stream_id, frame_datas, queue_response):
    # This function can be removed once
    # the issue it's solving is resolved.
    # See to function's doc string for more info
    set_mock_aiko_messager()

    if isinstance(frame_datas, dict):
        frame_datas = [frame_datas]

    stream_parameters: dict = parse_stream_parameters(
        agent.pipeline,
        agent.pipeline_definition,
    )

    agent.pipeline.create_stream(stream_id, parameters=stream_parameters, queue_response=queue_response)
    for frame_id, frame in enumerate(frame_datas):
        stream = {
            "stream_id": stream_id,
            "frame_id": frame_id,
        }

        data_files = [
            DataFile(
                file_id=uuid4(),
                content=frame["content"],
                media_frame_index=0,
                content_type="text",
            )
        ]
        agent.pipeline.process_frame(stream, {"data_files": data_files})
    agent.pipeline.destroy_stream(stream_id)


DEFAULT_FILEPATH_SEPERATOR = "\n"
DEFAULT_CONTENT_SEPERATOR = b"===END=="


@agent_group.command("start")
@click.option(
    "--seperator",
    "-p",
    type=str,
    default=None,
    help="If --expect-filepaths is true the default is '\\n'. Else the the unix file seperator '{DEFAULT_CONTENT_SEPERATOR}'. This parameter is only used for piped inputs, if passing paths directly use spaces to separate paths",
)
@click.option("--expect-filepaths", "-f", is_flag=True, default=False)
@click.option("--step-task-ids", "-t", type=str, default=None, help="comma separate for multiple")
@click.option("--step-id", "-i", type=str, default=None)
@click.option("--stream-id", "-s", type=str, default=DEFAULT_STREAM_ID)
@click.option("--dump-definition", type=str, default=None)
@click.option("--allow-non-machine-user", is_flag=True, default=False)
@click.argument("agent_definition", type=click.Path(dir_okay=False, exists=False))
@click.argument("input_data", nargs=-1, type=click.STRING, required=False)
@click.pass_context
def _start(
    ctx,
    seperator,
    expect_filepaths,
    step_task_ids,
    step_id,
    stream_id,
    dump_definition,
    allow_non_machine_user,
    agent_definition,
    input_data,
):
    """Start a local Highlighter Agent to process data either from your local machine or from Highlighter tasks.

    When processing local files, a single stream is created to process all files.
    The Agent definition must have its first element as a
    DataSourceCapability, such as ImageDataSource, VideoDataSource,
    TextDataSource, JsonArrayDataSource, etc. The examples below assume
    the use of ImageDataSource.

    When processing Highlighter tasks, a single stream is created for each
    task. The Agent definition should use AssessmentRead as the first element in this case.
    Note: When processing tasks, use a GraphQL API key specific to the agent being run. You can
    create this using 'hl agent create-token'.

    Examples:

      \b
      1. Start an agent against a single image path
      \b
        > hl agent start -f agent-def.json images/123.jpg

      \b
      2. Start an agent against a multiple image paths
      \b
        > find images/ -name *.jpg | hl agent start -f agent-def.json

      \b
      3. Cat the contents of an image to an agent
      \b
        > cat images/123.jpg | hl agent start -f agent-def.json

      \b
      4. Pass data directly to process_frame
      \b
        > hl agent start -f agent-def.json '[{"foo": "bar"},{"foo": "baz"}]'

      \b
      5. Process tasks from a Highlighter machine-step, using a local agent definition
      \b
        > STEP_UUID=xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
        > hl agent start agent-def.json --step-id "$STEP_UUID"

      \b
      6. Process tasks from a Highlighter machine-step, using an agent definition from Highlighter
      \b
        > STEP_UUID=xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
        > AGENT_UUID=xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
        > hl agent start "$AGENT_UUID" --step-id "$STEP_UUID"

    """
    if not input_data:
        input_data = "--"

    if (seperator is None) and (expect_filepaths):
        seperator = DEFAULT_FILEPATH_SEPERATOR
    elif (seperator is None) and (not expect_filepaths):
        seperator = DEFAULT_CONTENT_SEPERATOR

    if step_id and step_task_ids:
        raise ValueError()

    agent = HLAgent(agent_definition, dump_definition=dump_definition)
    agent.run_in_thread()

    # This should only be populated if running via the HLAgentCliRunner
    _hlagent_cli_runner_queue_response = ctx.obj.get("queue_response", None)

    data_sources = None

    if expect_filepaths:

        data_sources = _read_filepaths(input_data, seperator, "utf-8")

    elif _reading_raw_data_from_stdin_buffer(
        input_data, expect_filepaths, seperator
    ) and not _data_sources_are_specified_in_agent_definition(agent):
        """When reading raw data from the stdin buffer we need to:

        - use the `pipe://` scheme
        - determine the datatype expected by the head DataSource<Capability|PipelineElement>
        - ? force the DataSource<Capability|PipelineElement> to deal with the splitting of the
          buffer into individual files, or do the split here, not sure yet
        """
        data_sources = ["hlpipe://"]

    if step_id:
        agent.poll_for_tasks_loop(step_id, allow_non_machine_user=allow_non_machine_user)
    elif step_task_ids:
        client = ctx.obj["client"]
        for task_id in [t.strip() for t in step_task_ids.split(",")]:
            task = lease_task(
                client,
                task_id=task_id,
                lease_sec=TASK_LEASE_DURATION_SECONDS,
                set_status_to="RUNNING",
            )
            agent._process_task(task)
    elif data_sources or _data_sources_are_specified_in_agent_definition(agent):
        process_data_sources(agent, stream_id, data_sources, _hlagent_cli_runner_queue_response)
    else:
        # assume process_frame_data is passed in directly either as a json
        # str in input_data or via sdtin buffer
        # assert False, f"-------------: {input_data}"
        try:
            if input_data == "--":
                frame_datas = json.load(sys.stdin.buffer)
            else:
                frame_datas = json.loads(input_data[0])
        except Exception as e:
            raise ValueError(f"{e} -- {input_data}")
        loop_over_process_frame(agent, stream_id, frame_datas, _hlagent_cli_runner_queue_response)

    agent.stop()


@agent_group.command("create-token")
@click.option("--machine-agent-version-id", type=str)
@click.option("--machine-agent-name", type=str, required=False)
@click.option("--machine-agent-version-name", type=str, required=False)
def _create_token(
    machine_agent_version_id,
    machine_agent_name,
    machine_agent_version_name,
):
    """Create an access token for an agent

    Once an access token has been created, run an agent with that identity by
    setting `HL_WEB_GRAPHQL_API_TOKEN=<new-token>` before running `hl agent start`

    Either provide the ID of the machine-agent-version in Highlighter, or
    specify a new machine-agent name and version-name to create a new machine-agent-version
    for your agent.
    """
    if machine_agent_version_id is None:
        if machine_agent_name is not None or machine_agent_version_name is not None:
            raise ValueError(
                "Must specify either 'machine_agent_token', give a machine-agent version "
                "ID as the agent definition, or specify both 'machine_agent_name' and "
                "'machine_agent_version'"
            )
        machine_agent_version = create_machine_agent_version(machine_agent_name, machine_agent_version_name)
        machine_agent_version_id = machine_agent_version.id
    machine_agent_token = create_agent_token(machine_agent_version_id)
    # Print to stdout rather than log
    print(machine_agent_token)



# Usage
# ~~~~~
# connection = Connection(id)
# try:
#     connection.lock_acquire("function_name()")
#     ## Critial code ##
# finally:
#     connection.lock_release()
#
# Notes
# ~~~~~
# This is a low-level mechanism, used by the AikoServices framework
# and should not have any static dependencies on that framework
#
# Must use a Lock for "self.connection_state" updates via concurrent Threads.
# See "process.py", where both the main thread and the MQTT thread may
# independently perform simultaneous ConnectionState updates as the
# MQTT server and Registrar are both discovered and lost over time.
#
# To Do
# ~~~~~
# - Ensure all updates occur via "events" (handlers, messages and timers)
#   managed by the event loop and executed solely by the event loop thread

from aiko_services.main.utilities.lock import Lock

__all__ = ["ConnectionState", "Connection"]

class ConnectionState:
    NONE = "NONE"
    NETWORK = "NETWORK"      # Wi-Fi or Ethernet available
    BOOTSTRAP = "BOOTSTRAP"  # MQTT configuration found
    TRANSPORT = "TRANSPORT"  # MQTT, Ray, ROS2, ZeroMQ, etc
    REGISTRAR = "REGISTRAR"  # Registrar available for use

    states = [NONE, NETWORK, TRANSPORT, REGISTRAR]  # order matters

    @classmethod
    def index(cls, connection_state):  # throws ValueError
        return cls.states.index(connection_state)

class Connection:
    def __init__(self, id):
        self.id = id
        self.lock = Lock(f"connection.{id}")
        self.connection_state = ConnectionState.NONE
        self.connection_state_handlers = []

    def add_handler(self, connection_state_handler):
        connection_state_handler(self, self.connection_state)
        if not connection_state_handler in self.connection_state_handlers:
            self.connection_state_handlers.append(connection_state_handler)

    def get_state(self):
        return self.connection_state

    def is_connected(self, connection_state):  # throws ValueError
        current_state_index = ConnectionState.index(self.connection_state)
        return current_state_index >= ConnectionState.index(connection_state)

    def lock_acquire(self, location):
        self.lock.acquire(location)

    def lock_release(self):
        self.lock.release()

    def remove_handler(self, connection_state_handler):
        if connection_state_handler in self.connection_state_handlers:
           self.connection_state_handlers.remove(connection_state_handler)

    def update_state(self, connection_state, logger=None):
        if logger:
            if not self.lock.in_use():
                diagnostic = "update_state() invoked, whilst not locked"
                logger.error(f'Connection "{self.id}": {diagnostic}')

            state_change = f"{self.connection_state} --> {connection_state}"
            logger.info(f"Connection state: {state_change}")

        self.connection_state = connection_state

        for connection_state_handler in self.connection_state_handlers:
            connection_state_handler(self, connection_state)



# Usage
# ~~~~~
# pytest [-s] unit/test_pipeline_graph.py
# pytest [-s] unit/test_pipeline_graph.py::test_graph_name_mapping
#
# test_graph_name_mapping()
# -
#   See https://github.com/geekscape/aiko_services/pull/27
#
# To Do
# ~~~~~
# - Test all sorts of Pipeline Graphs

from typing import Tuple

import aiko_services as aiko
from aiko_services.tests.unit import do_create_pipeline

# PIPELINE_DEFINITION = HEADER + GRAPHS[PIPELINE_GRAPH_INDEX] + ELEMENTS
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

PIPELINE_GRAPH_INDEX = 0

HEADER = """{ "version": 0, "name": "p_test", "runtime": "python",
"""

GRAPHS = [
#### GRAPH 0 #### Helpful milestone ####
""" "graph": ["(A B (a_out_0: b_in_0 a_out_1: b_in_1) Terminate)"], """,
#### GRAPH 1 #### Helpful milestone ####
""" "graph": ["(A B (a_out_0: b_in_0 a_out_1: b_in_1) C Terminate)"], """,
#### GRAPH 2 #### Helpful milestone ####
""" "graph": ["(A B (a_out_0: b_in_0 a_out_1: b_in_1) C (a_out_0: c_in_0 b_out_0: c_in_1 a_out_1: c_in_2) Terminate)"], """,
#### GRAPH 3 #### Helpful milestone ####
""" "graph": ["(A B (A.a_out_0: b_in_0 A.a_out_1: b_in_1) Terminate)"], """,
#### GRAPH 4 #### Final test ####
""" "graph": ["(A B (A.a_out_0: b_in_0 A.a_out_1: b_in_1) C (A.a_out_0: c_in_0 B.b_out_0: c_in_1 A.a_out_1: c_in_2) Terminate)"], """
]

ELEMENTS = """
  "elements": [
    { "name":   "A",
      "input":  [],
      "output": [{ "name": "a_out_0", "type": "int" },
                 { "name": "a_out_1", "type": "int" }],
      "deploy": {
        "local": { "module": "aiko_services.tests.unit.test_pipeline_graph" }
      }
    },
    { "name":   "B",
      "input":  [{ "name": "b_in_0", "type": "int" },
                 { "name": "b_in_1", "type": "int" }],
      "output": [{ "name": "b_out_0", "type": "List[int]" }],
      "deploy": {
        "local": { "module": "aiko_services.tests.unit.test_pipeline_graph" }
      }
    },
    { "name":   "C",
      "input":  [{ "name": "c_in_0", "type": "int" },
                 { "name": "c_in_1", "type": "List[int]" },
                 { "name": "c_in_2", "type": "int"}],
      "output": [{ "name": "c_out_0", "type": "List[int]" }],
      "deploy": {
        "local": { "module": "aiko_services.tests.unit.test_pipeline_graph" }
      }
    },
    { "name":   "Terminate", "input":  [], "output": [],
      "deploy": {"local": { "module": "aiko_services.tests.unit.common" }}
    }
  ]
}
"""

class A(aiko.PipelineElement):
    def __init__(self, context):
        context.get_implementation("PipelineElement").__init__(self, context)

    def process_frame(self, stream):
        a_out_0 = 0
        a_out_1 = 1
        self.logger.info(
            f"{self.my_id()}: a_out_0: {a_out_0}, a_out_1: {a_out_1}")
    #   result = {"a_out_0": a_out_0, "a_out_1": a_out_1}
        result = {"a_out_0": a_out_0, "a_out_1": a_out_1, "b_in_1": -1}
        return aiko.StreamEvent.OKAY, result

class B(aiko.PipelineElement):
    def __init__(self, context):
        context.get_implementation("PipelineElement").__init__(self, context)

    def process_frame(self, stream, b_in_0, b_in_1):
        b_out_0 = [b_in_0, b_in_1]
        self.logger.info(f"{self.my_id()}: b_out_0: {b_out_0}")
    #   result = {"b_out_0": b_out_0}
        result = {"b_out_0": b_out_0, "c_in_0": -1, "c_in_1": [-1], "c_in_2": -1}
        return aiko.StreamEvent.OKAY, result

class C(aiko.PipelineElement):
    def __init__(self, context):
        context.get_implementation("PipelineElement").__init__(self, context)

    def process_frame(self, stream, c_in_0, c_in_1, c_in_2):
        c_out_0 = [c_in_0] + c_in_1 + [c_in_2]
        self.logger.info(f"{self.my_id()}: c_out_0: {c_out_0}")
        return aiko.StreamEvent.OKAY, {"c_out_0": c_out_0}

def test_graph_name_mapping():
    for graph_index in range(len(GRAPHS)):
        print("\n=================================")
        print(f"Test Graph name mapping: Graph[{graph_index}]")
        print("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")

        pipeline_definition = HEADER + GRAPHS[graph_index] + ELEMENTS
        do_create_pipeline(pipeline_definition)
        break



from typing import Dict, List, Optional, Tuple, Union

from highlighter.agent.capabilities import Capability, StreamEvent
from highlighter.client import DataFile
from highlighter.core import LabeledUUID

__all__ = ["{{cookiecutter.capability_class_name}}"]

ATTRIBUTE_UUID = LabeledUUID(int=2, label="response")

class {{cookiecutter.capability_class_name}}(Capability):
    """Does something cool
    """

    def __init__(self, context):
        context.get_implementation("PipelineElement").__init__(self, context)

    def start_stream(self, stream, stream_id) -> Tuple[StreamEvent, Optional[str]]:
        return StreamEvent.OKAY, None

    {% if cookiecutter.data_type == 'text' -%}
    def process_frame(self, stream, data_files: List[DataFile]) -> Tuple[StreamEvent, Union[Dict, str]]:
        for df in data_files:
            self.logger.info(f"DataFile.content: {df.content}")
        return StreamEvent.OKAY, {}
    {% elif cookiecutter.data_type == 'video' -%}
    def process_frame(self, stream, data_files: List[DataFile]) -> Tuple[StreamEvent, Union[Dict, str]]:
        for df in data_files:
            self.logger.info(f"DataFile.content: {df.content.shape}")
        return StreamEvent.OKAY, {}
    {% elif cookiecutter.data_type == 'image' -%}
    def process_frame(self, stream, data_files: List[DataFile]) -> Tuple[StreamEvent, Union[Dict, str]]:
        for df in data_files:
            self.logger.info(f"DataFile.content: {df.content.shape}")
        return StreamEvent.OKAY, {}
    {%- endif -%}






from typing import List

from ...core import GQLBaseModel
from .page_info import PageInfo

__all__ = ["ObjectClass", "ObjectClassTypeConnection"]


class ObjectClass(GQLBaseModel):
    id: str
    uuid: str
    name: str


class ObjectClassTypeConnection(GQLBaseModel):
    page_info: PageInfo
    nodes: List[ObjectClass]



import asyncio
import os

from tqdm import tqdm

__all__ = [
    "DEFAULT_PAGE_SIZE",
    "paginate",
]

DEFAULT_PAGE_SIZE = int(os.environ.get("HCP_PAGINATE_PAGE_SIZE", "200"))


def paginate(
    fn,
    return_type,
    page_size=DEFAULT_PAGE_SIZE,
    **kwargs,
):
    """
    fn: Graphql query that supports pagination
    return_type: Must contain be a BaseModel that contains

        page_info: PageInfo
        nodes: List[...]

    Where List[...] is a list of the type of the thing you want
    to return. EG

    class DatasetAssessment(BaseModel):
      stuff: ...

    class DatasetSubmissionConnection(BaseModel):
      page_info: PageInfo
      nodes: List[DatasetAssessment]

    client = HLClient...
    generator = paginate(
        client.datasetSubmissionConnection,
        DatasetSubmissionConnection,
        datasetId=123,
        )

    """

    def get_response(**kwargs):
        try:
            response = fn(
                return_type=return_type,
                first=page_size,
                **kwargs,
            )
            pbar.update(page_size)
        except asyncio.exceptions.TimeoutError:
            message = f"This could be becuase page_size ({page_size}) is too large"
            raise asyncio.exceptions.TimeoutError(message)
        return response

    with tqdm(desc=return_type.__name__) as pbar:
        response = get_response(**kwargs)

        for node in response.nodes:
            yield node

        while response.page_info.has_next_page:
            kwargs["after"] = response.page_info.end_cursor
            response = get_response(**kwargs)

            for node in response.nodes:
                yield node



#!/usr/bin/env python3
#
# Aiko Service: Recorder
# ~~~~~~~~~~~~~~~~~~~~~~
#
# Usage
# ~~~~~
# mosquitto_sub -t '#' -v
# REGISTRAR=0 registrar &
# RECORDER=0 ./recorder.py [topic_path_filter] &
#
# Where "topic_path_filter" default: "aiko/+/+/+/log"
#
# To Do
# ~~~~~
# - Improve CLI to record multiple different topic paths
# - On-the-fly configuration parameter updates ...
#   - _RING_BUFFER_SIZE, _TOPIC_LRU_CACHE_SIZE
#   - topic_path_filter causes unsubscribe and resubscribe to correct topic
# - Keep statistics for ...
#   - Topic LRU cache length
#   - Total messages received / sent, messages received / sent per second
# - Why doesn't Python MQTT client subscribe("+/+/+/+/log") work ?

import click
from collections import deque

from aiko_services.main import *
from aiko_services.main.utilities import *

_VERSION = 0

SERVICE_TYPE = "recorder"
PROTOCOL = f"{SERVICE_PROTOCOL_AIKO}/{SERVICE_TYPE}:{_VERSION}"

_LOGGER = aiko.logger(__name__)

_LRU_CACHE_SIZE = 2    # 128
_RING_BUFFER_SIZE = 2  # 128

# --------------------------------------------------------------------------- #

class Recorder(Service):
    Interface.default("Recorder", "aiko_services.main.recorder.RecorderImpl")

#   @abstractmethod
#   def METHOD_NAME(self):
#       pass

class RecorderImpl(Recorder):
    def __init__(self, context, topic_path_filter):
        context.get_implementation("Service").__init__(self, context)

# TODO: Add LRUCache popitem() handler to remove oldest ring buffer ?
#       And send ECProducer.remove(topic) to update the ECConsumer
        self.lru_cache = LRUCache(_LRU_CACHE_SIZE)

        self.share = {
            "lifecycle": "ready",
            "log_level": get_log_level_name(_LOGGER),
            "source_file": f"v{_VERSION}⇒ {__file__}",
            "lru_cache": {},                                             # HACK
            "lru_cache_size": _LRU_CACHE_SIZE,
            "ring_buffer_size": _RING_BUFFER_SIZE,
            "topic_path_filter": topic_path_filter
        }
        self.ec_producer = ECProducer(self, self.share)
        self.ec_producer.add_handler(self._ec_producer_change_handler)

        self.add_message_handler(self.recorder_handler, topic_path_filter)

    def _ec_producer_change_handler(self, command, item_name, item_value):
        if item_name == "log_level":
            _LOGGER.setLevel(str(item_value).upper())

    def recorder_handler(self, aiko, topic, payload_in):
        if topic in self.lru_cache:
            ring_buffer = self.lru_cache.get(topic)
        else:
            ring_buffer = deque(maxlen=_RING_BUFFER_SIZE)
# TODO: If LRUCache popitem(), then manually remove oldest ring buffer ?
#       And send ECProducer.remove(topic) to update the ECConsumer
            self.lru_cache.put(topic, ring_buffer)

# TODO: "utilities/parser.py": generate() and parse() need to handle
#       log messages with special characters ... use Canonical S-Expressions ?

        log_record = payload_in.replace(" ", " ")  # Unicode U00A0 (NBSP)
        log_record = log_record.replace("(", "{")
        log_record = log_record.replace(")", "}")
        ring_buffer.append(log_record)
        self.ec_producer.update(f"lru_cache.{topic}", log_record)        # HACK

# TODO: "share.py:ECConsumer._consumer_handler()" needs to handle list and dict
#       Appears that the "(add ...)" fails, but "(update ...)" works ?
#       Dashboard being updated with dict of entries that are lists ... works !

# --------------------------------------------------------------------------- #

@click.command("main", help="Recorder Service")
@click.argument("topic_path_filter", nargs=1, required=False,
    default=f"{get_namespace()}/+/+/+/log")

def main(topic_path_filter):
    tags = ["ec=true"]  # TODO: Add ECProducer tag before add to Registrar
    init_args = service_args(SERVICE_TYPE, None, None, PROTOCOL, tags)
    init_args["topic_path_filter"] = topic_path_filter
    recorder = compose_instance(RecorderImpl, init_args)
    aiko.process.run()

if __name__ == "__main__":
    main()

# --------------------------------------------------------------------------- #



import os
import shutil
from pathlib import Path
from tempfile import TemporaryDirectory

import click
from cookiecutter.generate import generate_context
from cookiecutter.main import cookiecutter
from cookiecutter.prompt import prompt_for_config

from highlighter.client import TrainingConfigType
from highlighter.client.gql_client import HLClient
from highlighter.core.logging import ColourStr
from highlighter.datasets.dataset import Dataset
from highlighter.trainers import _scaffold


@click.command("new")
@click.argument("dest", type=click.Path(dir_okay=True, file_okay=False))
@click.pass_context
def new_cmd(ctx, dest):
    """Create a new Highlighter scaffold directory

    \b
    DEST/
      TITLE/
        highlighter/
          context.json
        pyproject.toml
        README.md
        shell.nix

    """
    dest = Path(dest)
    if not dest.exists():
        dest.mkdir(parents=True)

    template_path = Path(__file__).parent.parent / "templates" / "scaffold"

    # Load the context from the Cookiecutter template
    context = generate_context(context_file=template_path / "cookiecutter.json")
    # Prompt user for inputs
    completed_context = prompt_for_config(context)

    _final_path = dest / completed_context["title_slug"]
    if _final_path.exists():
        raise ValueError(f"{_final_path.absolute()} is already a scaffold")

    final_path = cookiecutter(
        str(template_path), no_input=True, extra_context=completed_context, output_dir=str(dest)
    )

    _scaffold.append_hl_scaffold_context(
        Path(final_path), {"cookiecutter_scaffold_context": completed_context}
    )
    click.echo(f"Project created at {final_path}")


@click.group("generate")
@click.pass_context
def generate_group(ctx):
    pass


def _append_readmes(scaffold_dir, tmp_agent_root):
    with (scaffold_dir / "README.md").open("r") as f:
        scaffold_readme = f.read()

    if not "Run Agent" in scaffold_readme:
        with (tmp_agent_root / "README.md").open("r") as f:
            agent_readme = f.read()

        scaffold_readme = f"{scaffold_readme}\n\n{agent_readme}"

        with (scaffold_dir / "README.md").open("w") as f:
            f.write(scaffold_readme)


def _append_capability_inits(src_init, new_init):
    with new_init.open("r") as f:
        new = f.read()

    if src_init.exists():
        with src_init.open("a") as f:
            f.write(new)
    else:
        src_init.parent.mkdir(exist_ok=True)
        with src_init.open("w") as f:
            f.write(new)


def _move_tmp_dirs(scaffold_dst_dir, tmp_agent_src_dir):
    scaffold_dst_dir.mkdir(exist_ok=True)

    for p in os.listdir(str(tmp_agent_src_dir)):
        dst = scaffold_dst_dir / p
        if dst.exists():
            continue
        src = tmp_agent_src_dir / p
        if Path(src).is_file():
            shutil.copy2(src, dst)


def _remove(filepath):
    print(f"Removing: {filepath}")
    if os.path.isfile(filepath):
        os.remove(filepath)
    elif os.path.isdir(filepath):
        shutil.rmtree(filepath)


def _cleanup_test_inputs(scaffold_dst_dir, data_type):
    test_data_files = [
        ("image", "test.png"),
        ("video", "test.mp4"),
        ("text", "test.txt"),
    ]

    for dt, name in test_data_files:
        if data_type == dt:
            continue

        _remove(scaffold_dst_dir / name)


@generate_group.command("agent")
@click.argument("scaffold_dir", type=click.Path(dir_okay=True, file_okay=False))
@click.pass_context
def agent_generate(ctx, scaffold_dir):
    scaffold_dir = Path(scaffold_dir)
    highlighter_dir = _scaffold.DIRS.hl_cache(scaffold_dir)
    if not highlighter_dir.exists():
        raise ValueError(
            f"DEST '{scaffold_dir}' must be an existing highlighter scaffold with a highlighter directory. Create one by running 'hl new'"
        )
    hl_ctx = _scaffold.load_hl_context(_scaffold.DIRS.hl_context_json(scaffold_dir))

    template_path = Path(__file__).parent.parent / "agent" / "template"

    # Load the context from the Cookiecutter template
    context = generate_context(context_file=template_path / "cookiecutter.json")
    # Prompt user for inputs
    completed_context = prompt_for_config(context)
    scaffold_pkg_name = hl_ctx["cookiecutter_scaffold_context"]["title_slug"]
    completed_context["_scaffold_pkg_name"] = scaffold_pkg_name

    with TemporaryDirectory() as tmp:

        final_path = cookiecutter(
            str(template_path),
            no_input=True,
            extra_context=completed_context,
            output_dir=tmp,
        )

        _append_readmes(scaffold_dir, Path(final_path))
        _append_capability_inits(
            scaffold_dir / "src" / scaffold_pkg_name / "capabilities" / "__init__.py",
            Path(final_path) / "capabilities" / "__init__.py",
        )
        _move_tmp_dirs(
            scaffold_dir / "src" / scaffold_pkg_name / "capabilities", Path(final_path) / "capabilities"
        )

        _move_tmp_dirs(scaffold_dir / "agents", Path(final_path) / "agents")

        _move_tmp_dirs(scaffold_dir / "inputs", Path(final_path) / "inputs")


@generate_group.command("training-run")
@click.argument("training_run_id", type=int)
@click.argument("trainer", type=click.Choice(list(_scaffold.TrainerType)))
@click.argument("scaffold_dir", type=click.Path(dir_okay=True, file_okay=False))
@click.option("--yes", "-y", is_flag=True, default=False)
@click.pass_context
def train_generate(ctx, training_run_id, trainer, scaffold_dir, yes):
    scaffold_dir = Path(scaffold_dir)
    highlighter_dir = _scaffold.DIRS.hl_cache(scaffold_dir)
    if not highlighter_dir.exists():
        raise ValueError(
            f"DEST '{scaffold_dir}' must be an existing highlighter scaffold with a highlighter directory"
        )

    client: HLClient = ctx.obj["client"]
    highlighter_training_config = TrainingConfigType.from_highlighter(client, training_run_id)

    training_run_dir = _scaffold.DIRS.training_run_dir(scaffold_dir, str(training_run_id))
    hl_training_config_json = _scaffold.DIRS.hl_training_config(scaffold_dir, str(training_run_id))
    if training_run_dir.exists() and (yes or _scaffold.ask_to_overwrite(training_run_dir, terminate=True)):
        click.echo(f"Removing '{training_run_dir}'")
        shutil.rmtree(training_run_dir)
        click.echo(f"Removing '{hl_training_config_json.parent}'")
        shutil.rmtree(hl_training_config_json.parent)

    training_run_dir.mkdir(parents=True, exist_ok=False)
    hl_training_config_json.parent.mkdir(parents=True, exist_ok=False)

    highlighter_training_config.dump("json", hl_training_config_json)

    dataset_cache_dir = _scaffold.DIRS.hl_training_run_cache_datasets_dir(
        scaffold_dir,
        str(training_run_id),
    )
    dataset_cache_dir.mkdir()

    Dataset.read_training_config(client, highlighter_training_config, dataset_cache_dir)

    if "yolo" in trainer:
        from highlighter.trainers.yolov11 import generate
    else:
        raise ValueError(f"Unable to determine trainer from '{trainer}'")

    _scaffold.append_hl_scaffold_context(
        scaffold_dir, {f"training_run_{training_run_id}": {"trainer": trainer}}
    )

    generate(training_run_dir, trainer)
    click.echo(f"{trainer} template generated at {training_run_dir}")
    cmd = ColourStr.green(f"hl train start {training_run_dir}")
    if scaffold_dir.absolute() == Path.cwd():
        click.echo(f"Next, run `{cmd}`")
    else:
        cd_cmd = ColourStr.green(f"cd {scaffold_dir}")
        click.echo(f"Next, `{cd_cmd}` and run `{cmd}`")



from typing import Dict, Optional, Tuple
from uuid import UUID

from highlighter import Entity
from highlighter.agent.capabilities.base_capability import Capability, StreamEvent


class GroupEntities(Capability):
    """
    Capability for grouping entities according to the enum values of a particular taxonomy attribute.

    Example element definition:
    {
        "name": "Group Entities",
        "input": [{ "name": "entities", "type": "bytes" }],
        "output": [{ "name": "cat", "type": "bytes" },
                   { "name": "tardigrade", "type": "bytes" }],
        "deploy": { "local": { "module": "highlighter.agent.capabilities.group_entities",
                               "class_name": "GroupEntities" } },
        "parameters": {
            "enumIdToOutputName": {
                "91e7f435-001c-4638-aed0-e067eded9fa7": "tardigrade",
                "a7fe0c3c-92c6-44fe-85cf-21df02a02326": "cat"
            },
            "groupByAttributeId": "df10b67d-b476-4c4d-acc2-c1deb5a0e4f4",
            "groupByAttributeName": "object"
        }
    }

    """

    class DefaultStreamParameters(Capability.DefaultStreamParameters):
        groupByAttributeId: Optional[UUID] = None
        enumIdToOutputName: Dict[UUID, str] = {}

    def process_frame(self, stream, entities: Dict[UUID, Entity]) -> Tuple[StreamEvent, Dict]:
        group_by_attribute_id, _ = self._get_parameter("group_by_attribute_id")
        group_by_attribute_id = UUID(group_by_attribute_id)
        enum_id_to_output_name, _ = self._get_parameter("enum_id_to_output_name")
        grouped_entities: Dict[str, Dict[UUID, Entity]] = {
            output_name: {} for output_name in enum_id_to_output_name.values()
        }
        grouped_entity_ids = []
        for entity_id, entity in entities.items():
            for annotation in entity.annotations:
                for observation in annotation.observations:
                    if observation.attribute_id == group_by_attribute_id:
                        output_name = enum_id_to_output_name[str(observation.value)]
                        grouped_entities[output_name][entity_id] = entity
                        grouped_entity_ids.append(entity_id)
        if len(grouped_entity_ids) < len(entities):
            return StreamEvent.ERROR, {"diagnostic": "Entities were not all able to be grouped"}
        return StreamEvent.OKAY, grouped_entities



from datetime import datetime, timezone
from typing import Any, Dict, List, Optional, Tuple, Union
from uuid import UUID, uuid4

import numpy as np
import shapely.geometry as geom
from PIL.Image import Image
from pydantic import BaseModel, ConfigDict, Field, field_validator

from highlighter.core.hl_base_model import HLModelMap

from ...core import (
    OBJECT_CLASS_ATTRIBUTE_UUID,
    BelongsTo,
    HasMany,
    HLDataModel,
    polygon_from_mask,
)
from .base_models import EAVT
from .data_file import DataFile
from .datum_source import DatumSource
from .observation import Observation

__all__ = ["Annotation"]


class AnnotationCrop(BaseModel):
    content: Any
    annotation_id: UUID
    entity_id: Optional[UUID] = None


@HasMany("observations", target_cls="highlighter.client.Observation", back_populates="annotation")
@BelongsTo(
    "entity",
    target_cls="highlighter.client.Entity",
    back_populates="annotations",
)
class Annotation(HLDataModel):
    id: UUID = Field(..., default_factory=uuid4)

    location: Optional[Union[geom.Polygon, geom.MultiPolygon, geom.LineString, geom.Point]] = None
    data_file_id: Optional[UUID] = None
    observations: HLModelMap = Field(..., default_factory=lambda: HLModelMap(Observation))
    entity_id: Optional[UUID] = None

    # TODO update HL Web to refer to data-sources (e.g. webrtc streams) as well as files
    track_id: Optional[UUID] = None
    correlation_id: Optional[UUID] = None

    datum_source: DatumSource

    model_config = ConfigDict(arbitrary_types_allowed=True)  # Required for shapely geometry types

    @field_validator("observations", mode="before")
    @classmethod
    def validate_c(cls, v: Any) -> HLModelMap:
        if isinstance(v, list):
            _v = HLModelMap(Observation)
        else:
            _v = v
        assert isinstance(_v, HLModelMap)
        return _v

    @classmethod
    def from_points(
        cls,
        points: List[Tuple[int, int]],
        confidence: float,
        data_file_id: Optional[UUID] = None,
        observations: Optional[List[Observation]] = None,
        entity_id: Optional[UUID] = None,
        track_id: Optional[UUID] = None,
        correlation_id: Optional[UUID] = None,
        # Optional DatumSource fields
        frame_id: Optional[int] = None,
        host_id: Optional[str] = None,
        pipeline_element_name: Optional[str] = None,
    ):
        datum_source = DatumSource(
            confidence=confidence,
            frame_id=frame_id,
            host_id=host_id,
            pipeline_element_name=pipeline_element_name,
        )

        anno = cls(
            datum_source=datum_source,
            location=geom.Polygon(points),
            data_file_id=data_file_id,
            entity_id=entity_id,
            track_id=track_id,
            correlation_id=correlation_id,
        )
        if observations is not None:
            _ = [anno.observations.add(o) for o in observations]
        return anno

    @classmethod
    def from_left_top_right_bottom_box(
        cls,
        box: Tuple[int, int, int, int],
        confidence: float,
        data_file_id: Optional[UUID] = None,
        observations: Optional[List[Observation]] = None,
        entity_id: Optional[UUID] = None,
        track_id: Optional[UUID] = None,
        correlation_id: Optional[UUID] = None,
        # Optional DatumSource fields
        frame_id: Optional[int] = None,
        host_id: Optional[str] = None,
        pipeline_element_name: Optional[str] = None,
    ):
        x0, y0, x1, y1 = box
        return cls.from_points(
            ((x0, y0), (x1, y0), (x1, y1), (x0, y1), (x0, y0)),
            confidence,
            data_file_id=data_file_id,
            observations=observations,
            entity_id=entity_id,
            track_id=track_id,
            correlation_id=correlation_id,
            frame_id=frame_id,
            host_id=host_id,
            pipeline_element_name=pipeline_element_name,
        )

    @classmethod
    def from_mask(
        cls,
        mask: np.array,
        confidence: float,
        data_file_id: Optional[UUID] = None,
        observations: Optional[List[Observation]] = None,
        entity_id: Optional[UUID] = None,
        track_id: Optional[UUID] = None,
        correlation_id: Optional[UUID] = None,
        # Optional DatumSource fields
        frame_id: Optional[int] = None,
        host_id: Optional[str] = None,
        pipeline_element_name: Optional[str] = None,
        **polygon_from_mask_kwargs,
    ):

        points = polygon_from_mask(mask, **polygon_from_mask_kwargs)
        points = [(x, y) for x, y in zip(points[:-1:2], points[1::2])]
        # Shapely expects a closed
        points.append(points[0])
        # breakpoint()

        return cls.from_points(
            points,
            confidence,
            data_file_id=data_file_id,
            observations=observations,
            entity_id=entity_id,
            track_id=track_id,
            correlation_id=correlation_id,
            frame_id=frame_id,
            host_id=host_id,
            pipeline_element_name=pipeline_element_name,
        )

    def serialize(self):
        return {
            "id": str(self.id),
            "entity_id": str(self.entity_id),
            "location": self.location.wkt if self.location is not None else None,
            "observations": [o.serialize() for o in self.observations.values()],
            "track_id": str(self.track_id),
            "correlation_id": str(self.correlation_id),
            "data_file_id": str(self.data_file_id),
            "datum_source": self.datum_source.serialize(),
        }

    def get_observation(self, attribute_id: UUID) -> Optional[Observation]:
        for o in self.observations:
            if o.attribute_id == attribute_id:
                return o
        return None

    def has_observation(self, attribute_id: UUID, value: Optional[Any] = None) -> bool:
        if value is None:
            return any([o.attribute_id == attribute_id for o in self.observations])
        else:
            return any([((o.attribute_id == attribute_id) and (o.value == value)) for o in self.observations])

    def crop(self, crop_args: "CropArgs") -> AnnotationCrop:
        from ...datasets.cropping import crop_rect_from_poly

        if self.data_file_id is None:
            raise ValueError("Cannot crop an Annotation when data_file_id is None")
        if self.location is None:
            raise ValueError("Cannot crop an Annotation when location is None")
        if not isinstance(self.location, geom.Polygon):
            raise ValueError(f"Cannot crop an Annotation when location is {type(self.location)}")

        data_file = DataFile.find_by_id(self.data_file_id)

        if data_file is None:
            raise ValueError(f"Could not find DataFile with id {self.data_file_id}")

        if "image" not in data_file.content_type:
            raise ValueError(
                f"Cannot crop data_file with content_type '{data_file.content_type}', must be 'image'"
            )

        if not isinstance(data_file.content, (np.ndarray, Image)):
            raise ValueError(
                f"Cannot crop data_file with content '{type(data_file.content)}', must be (PIL.Image|np.array)"
            )

        cropped_image = crop_rect_from_poly(data_file.content, self.location, crop_args)
        return AnnotationCrop(content=cropped_image, entity_id=self.entity_id, annotation_id=self.id)

    def to_deprecated_pixel_location_eavt(self) -> EAVT:
        return EAVT.make_pixel_location_eavt(
            entity_id=self.entity_id,
            location_points=self.location,
            confidence=self.datum_source.confidence,
            time=datetime.now(timezone.utc),
            frame_id=self.datum_source.frame_id,
        )

    @field_validator("location")
    @classmethod
    def validate_geometry(cls, v):
        if v is not None:
            assert v.is_valid, f"Invalid Geometry: {v}"
        return v

    def to_json(self):
        data = self.model_dump()
        data["id"] = str(data["id"])
        data["entity_id"] = str(data["entity_id"])
        data["location"] = data["location"].wkt
        data["datum_source"] = data["datum_source"]
        data["observations"] = [d.to_json() for d in data["observations"].values()]

        data["track_id"] = str(data["track_id"]) if self.track_id is not None else None
        data["data_file_id"] = str(data["track_id"]) if self.track_id is not None else None
        data["correlation_id"] = str(data["track_id"]) if self.track_id is not None else None
        return data

    def gql_dict(self) -> Dict:
        try:
            object_class_observation = [
                observation
                for observation in self.observations
                if observation.attribute_id == OBJECT_CLASS_ATTRIBUTE_UUID
            ][-1]
        except IndexError:
            raise ValueError(
                "Annotation must have an object-class observation in order to submit to Highlighter"
            )

        if isinstance(self.location, geom.Polygon):
            data_type = "polygon"
        elif isinstance(self.location, geom.LineString):
            data_type = "line"
        elif isinstance(self.location, geom.Point):
            data_type = "point"
        else:
            data_type = "polygon"
        result = {
            "objectClassUuid": str(object_class_observation.value),
            "location": self.location.wkt if self.location is not None else None,
            "confidence": self.datum_source.confidence,
            "dataType": data_type,
            "correlationId": str(self.correlation_id),
            "frameId": self.datum_source.frame_id,
            "trackId": str(self.track_id),
            "entityId": str(self.entity_id),
            "dataFileId": str(self.data_file_id),
            "uuid": str(self.id),
        }
        return result



from typing import List, Optional
from uuid import UUID

from ...core import GQLBaseModel
from .page_info import PageInfo

__all__ = ["Image", "ImagePresigned", "ImageConnection"]


class Image(GQLBaseModel):
    id: str
    uuid: UUID
    width: Optional[int] = None
    height: Optional[int] = None
    original_source_url: str
    mime_type: str


class ImagePresigned(Image):
    file_url_original: str


class ImageConnection(GQLBaseModel):
    page_info: PageInfo
    nodes: List[ImagePresigned]



import os
import shutil
from pathlib import Path

import click
import yaml

from highlighter.client import TrainingConfigType
from highlighter.client.gql_client import HLClient
from highlighter.core.logging import ColourStr
from highlighter.datasets.dataset import Dataset
from highlighter.trainers import _scaffold


@click.group("train")
@click.pass_context
def train_group(ctx):
    pass


def _validate_training_run_dir(training_run_dir: Path):
    trainer_py = training_run_dir / "trainer.py"
    if trainer_py.exists():
        return training_run_dir

    ml_training_dir = training_run_dir / "ml_training"
    training_run_dirs = list(ml_training_dir.glob("*"))
    if len(training_run_dirs) == 1:
        return training_run_dirs[0]

    raise ValueError(f"Invalid training_run_dir {training_run_dir}")


@train_group.command("start")
@click.argument("training-run-dir", required=False, default=".")
@click.option("--yes", "-y", is_flag=True, default=False)
@click.pass_context
def train_start(ctx, training_run_dir, yes):
    client: HLClient = ctx.obj["client"]

    training_run_dir = Path(training_run_dir).absolute()
    training_run_dir = _validate_training_run_dir(training_run_dir)
    training_run_id = training_run_dir.stem

    scaffold_dir = training_run_dir.parent.parent

    hl_training_config_json = _scaffold.DIRS.hl_training_config(scaffold_dir, training_run_id)
    highlighter_training_config = TrainingConfigType.from_json(hl_training_config_json)
    dataset_cache_dir = _scaffold.DIRS.hl_training_run_cache_datasets_dir(scaffold_dir, training_run_id)
    datasets = Dataset.read_training_config(client, highlighter_training_config, dataset_cache_dir)

    hl_ctx = _scaffold.load_hl_context(_scaffold.DIRS.hl_context_json(scaffold_dir))
    trainer_type = hl_ctx[f"training_run_{training_run_id}"]["trainer"]

    if "yolo" in trainer_type:
        from highlighter.trainers.yolov11 import prepare_datasets

        combined_ds = prepare_datasets(datasets)
    else:
        raise ValueError(f"Invalid trainer_type '{trainer_type}'")

    if (training_run_dir / "datasets" / "data.yaml").exists() and (
        yes or _scaffold.ask_to_overwrite(training_run_dir / "datasets")
    ):
        click.echo(f"Removing: {training_run_dir / 'datasets'}")
        shutil.rmtree(training_run_dir / "datasets")

    trainer = _scaffold.load_trainer_module(training_run_dir)

    os.chdir(training_run_dir)
    _, artefact_path = trainer.train(highlighter_training_config, combined_ds, training_run_dir)
    click.echo(f"Training {training_run_id} complete")
    cmd = ColourStr.green(
        f"hl training-run artefact create -i {training_run_id} -a {artefact_path.relative_to(training_run_dir)}"
    )
    click.echo(f"Next run: `{cmd}` to upload to Highlighter")


@train_group.command("export")
@click.argument("training-run-dir", required=False, default=".")
@click.argument("checkpoint", required=True, type=click.Path(dir_okay=False))
@click.argument("config", required=True, type=click.Path(dir_okay=False))
@click.pass_context
def train_export(ctx, training_run_dir, checkpoint, config):
    training_run_dir = Path(training_run_dir).absolute()
    training_run_dir = _validate_training_run_dir(training_run_dir)
    training_run_id = training_run_dir.stem

    scaffold_dir = training_run_dir.parent.parent

    hl_ctx = _scaffold.load_hl_context(_scaffold.DIRS.hl_context_json(scaffold_dir))
    trainer_type = hl_ctx[f"training_run_{training_run_id}"]["trainer"]

    trainer = _scaffold.load_trainer_module(training_run_dir)
    crop_args = getattr(trainer, "CROP_ARGS", None)

    if "yolo" in trainer_type:
        from highlighter.trainers.yolov11 import export

        artefact_path = export(checkpoint, config, training_run_dir, crop_args)
    else:
        raise ValueError(f"Unable to determine trainer from '{trainer_type}'")

    click.echo(f"Export {checkpoint} complete")
    cmd = ColourStr.green(
        f"hl training-run artefact create -i {training_run_id} -a {artefact_path.relative_to(Path.cwd())}"
    )
    click.echo(f"Next run: `{cmd}` to upload to Highlighter")



from datetime import datetime
from typing import List, Optional

from ..core import GQLBaseModel
from . import HLClient
from .base_models import MachineAgent, MachineAgentVersion, PipelineInstanceType


def create_machine_agent_version(
    machine_agent_name: str, machine_agent_version_name: str
) -> MachineAgentVersion:
    class CreateMachineAgentPayload(GQLBaseModel):
        errors: List[str]
        machine_agent: MachineAgent

    result = HLClient.get_client().createMachineAgent(
        return_type=CreateMachineAgentPayload,
        name=machine_agent_name,
        machineAgentTypeId="9efdcbce-3434-4b6c-b15b-1f3565899116",  # std_element_pipeline
    )
    if len(result.errors) > 0:
        raise ValueError(f"Error creating machine agent '{machine_agent_name}': {result.errors}")
    machine_agent_id = result.machineAgent.id

    class CreateMachineAgentVersionPayload(GQLBaseModel):
        errors: List[str]
        machine_agent_version: MachineAgentVersion

    result = HLClient.get_client().createMachineAgentVersion(
        return_type=CreateMachineAgentVersionPayload,
        title=machine_agent_version_name,
        code="",
        codeVersion="",
        machineAgentId=machine_agent_id,
        machineAgentTypeId="9efdcbce-3434-4b6c-b15b-1f3565899116",  # std_element_pipeline
    )
    if len(result.errors) > 0:
        raise ValueError(
            f"Error creating machine agent version '{machine_agent_version_name}': {result.errors}"
        )
    return result.machine_agent_version


def create_pipeline_instance(machine_agent_version_id: str, step_id: str) -> str:
    now = datetime.now().isoformat()

    class CreatePipelineInstance(GQLBaseModel):
        errors: List[str]
        pipeline_instance: PipelineInstanceType

    result = HLClient.get_client().createPipelineInstance(
        return_type=CreatePipelineInstance,
        stepId=step_id,
        machineAgentVersionId=machine_agent_version_id,
        reportedStatus="RUNNING",
        startedAt=now,
        lastReportedAt=now,
    )
    if len(result.errors) > 0:
        raise ValueError(f"Could not create pipeline instance in Highlighter: {result.errors}")
    return result.pipeline_instance.id


def update_pipeline_instance(pipeline_instance_id: str, status: str, message: Optional[str] = None):
    class UpdatePipelineInstance(GQLBaseModel):
        errors: List[str]

    result = HLClient.get_client().updatePipelineInstance(
        return_type=UpdatePipelineInstance,
        id=pipeline_instance_id,
        reportedStatus=status,
        lastReportedAt=datetime.now().isoformat(),
        message=message,
    )
    if len(result.errors) > 0:
        raise ValueError(f"Could not update pipeline instance in Highlighter: {result.errors}")


def create_agent_token(machine_agent_version_id: str) -> str:
    class CreateAccessTokenPayload(GQLBaseModel):
        class AccessTokenPayload(GQLBaseModel):
            token: str

        errors: List[str]
        access_token: AccessTokenPayload

    result = HLClient.get_client().createAccessToken(
        return_type=CreateAccessTokenPayload,
        machineAgentVersionId=str(machine_agent_version_id),
    )
    if len(result.errors) > 0:
        raise ValueError(f"Error creating access token for agent: {result.errors}")
    return result.access_token.token
